# -*- coding: utf-8 -*-
"""Global configuration"""

from configparser import ConfigParser
from collections import OrderedDict
from contextlib import suppress
from functools import wraps
import os

import click
import peewee as pw
from playhouse import db_url

from .ui import ClickUI


class DBProxy(pw.Proxy):
    """Database object sitting on top of Peewee"""
    OperationalError = pw.OperationalError

    def __init__(self, url=''):
        super().__init__()
        if url:
            self.load(url)

    def __call__(self, routine):
        @wraps(routine)
        def wrapper(*args, **kwargs):
            """decorator for routines needing database connection"""
            with self:
                retval = routine(*args, **kwargs)
            return retval

        return wrapper

    def __enter__(self):
        """context manager for routines needing database connection"""
        with suppress(pw.OperationalError):
            self.connect()
        return self

    def __exit__(self, *_):
        with suppress(pw.OperationalError):
            self.close()

    def load(self, url):
        """New database session"""
        try:
            base = db_url.connect(url)
            self.initialize(base)
        except RuntimeError:
            UI.display('Failed loading database at {}'.format(url))


class UIProxy(object):
    """UI abstraction layer"""
    impl = ClickUI()
    implementations = {'text_ui': ClickUI,
                       'click': ClickUI}

    def __init__(self, impl='click', verbosity=0):
        self.load(impl, verbosity)

    def __getattr__(self, name):
        result = getattr(self.impl, name)
        if result is None:
            raise NameError('{implementation} has no function named {function}'
                            .format(implementation=self.impl.__name__,
                                    function=name))
        else:
            return result

    def get_name(self):
        """Get the underlying user interface implementation's name."""
        return self.impl.__name__

    def load(self, implementation, verbosity):
        """Load another user interface implementation"""
        impl = self.implementations.get(implementation, ClickUI)
        self.impl = impl(verbosity)

USER_DATA_DIR = click.get_app_dir('rpi2caster', force_posix=True, roaming=True)
with suppress(FileExistsError):
    os.mkdir(USER_DATA_DIR, mode=0o775)

DEFAULTS = OrderedDict({'System': {}, 'Typesetting': {}})
DEFAULTS['System']['interfaces'] = ('http://localhost:23017, '
                                    'http://localhost:23017/interfaces/0, '
                                    'http://monotype:23017/interfaces/0')
DEFAULTS['System']['database'] = ('sqlite:////{}/rpi2caster.db'
                                  .format(USER_DATA_DIR))
DEFAULTS['System']['extra_wedges'] = 'extra_wedges.json'
DEFAULTS['System']['extra_languages'] = 'extra_languages.json'
DEFAULTS['System']['extra_unit_arrangements'] = 'extra_unit_arrangements.json'
DEFAULTS['System']['extra_typefaces'] = 'extra_typefaces.json'
DEFAULTS['Typesetting']['DEFAULTSault_measure'] = '25cc'
DEFAULTS['Typesetting']['measurement_unit'] = 'cc'

UI = UIProxy()
CFG = ConfigParser()
CFG.read_dict(DEFAULTS)
DB = DBProxy(CFG['System']['database'])



# -*- coding: utf-8 -*-
"""Contains functions used for justification, control sequences etc."""


def single_j(pos_0075=3, pos_0005=8, fwd=False,
             comment='Single justification'):
    """Add 0075 + pos_0075, then 0005 + pos_0005"""
    return (pump_start(pos_0075, comment) if pos_0075 == pos_0005
            else pump_stop(pos_0005) + pump_start(pos_0075, comment) if fwd
            else pump_start(pos_0075, comment) + pump_stop(pos_0005))


def double_j(pos_0075=3, pos_0005=8, fwd=False,
             comment='Double justification'):
    """Add 0075 + pos_0075, then 0005-0075 + pos_0005"""
    return (galley_trip(pos_0075, comment) if pos_0075 == pos_0005
            else galley_trip(pos_0005) + pump_start(pos_0075, comment) if fwd
            else pump_start(pos_0075, comment) + galley_trip(pos_0005))


def galley_trip(pos_0005=8, comment='Line to the galley'):
    """Put the line to the galley"""
    code = 'NKJS 0075 0005 {}'.format(pos_0005)
    return ['{:<20} // {}'.format(code, comment or '')]


def pump_start(pos_0075=3, comment='Starting the pump'):
    """Start the pump and set 0075 wedge"""
    code = 'NKS 0075 {}'.format(pos_0075)
    return ['{:<20} // {}'.format(code, comment or '')]


def pump_stop(pos_0005=8, comment='Stopping the pump'):
    """Stop the pump"""
    code = 'NJS 0005 {}'.format(pos_0005)
    return ['{:<20} // {}'.format(code, comment or '')]


def end_casting(fwd=False):
    """Alias for ending the casting job"""
    text1, text2 = 'End casting', 'Last line out'
    return (galley_trip(comment=text2) + pump_stop(comment=text1) if fwd
            else pump_stop(comment=text1) + galley_trip(comment=text2))



# -*- coding: utf-8 -*-
"""main_models - all big/database-dependent models for rpi2caster"""

from collections import defaultdict, OrderedDict
from contextlib import suppress
from itertools import chain, product
from functools import singledispatch
import json

import peewee as pw

from .rpi2caster import DB
from . import basic_models as bm, definitions as d, parsing as p, data


class BaseModel(pw.Model):
    """Base class for all models"""
    # define the class exception here, to appease linters
    DoesNotExist = pw.DoesNotExist

    class Meta:
        """Database metadata"""
        database = DB

        def db_table_func(self):
            """get a table name for a model"""
            try:
                model_name = self.__name__
            except AttributeError:
                model_name = self.__class__.__name__
            tables = dict(Diecase='matrix_cases', Ribbon='ribbons')
            return tables.get(model_name) or '{}s'.format(model_name.lower())


class Wedge:
    """Default S5-12E wedge, unless otherwise specified"""
    __slots__ = ('series', 'set_width', 'is_brit_pica', '_units')

    def __init__(self, wedge_name='', wedge_data=None):
        w_data = wedge_data or {}
        if wedge_name:
            self.name = wedge_name
        else:
            self.series = w_data.get('series', '5')
            self.set_width = w_data.get('set_width', 12.0)
            self.is_brit_pica = w_data.get('is_brit_pica', True)
            self._units = w_data.get('units', d.S5)

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<Wedge: {}>'.format(self.name)

    def __getitem__(self, row):
        return self.units[row]

    def __bool__(self):
        return True

    @property
    def pica(self):
        """Get the pica base of the wedge. Two values were used:
            .1667" (old British pica, European wedges), also: DTP pica,
            .1660" (new British, American), also: TeX pica."""
        return 0.1667 if self.is_brit_pica else 0.166

    @property
    def name(self):
        """Gets a wedge name - for example S5-12.25E.
        S (for stopbar) is prepended by convention.
        E is appended whenever the wedge is based on pica = .1667".
        """
        # Truncate the fractional part of the set width if it's an integer
        set_w = self.set_width if self.set_width % 1 else int(self.set_width)

        name = 'S{series}-{set_width}{european_suffix}'
        return name.format(series=self.series, set_width=set_w,
                           european_suffix=self.is_brit_pica * 'E')

    @name.setter
    def name(self, wedge_name):
        """Parse the wedge name to get series, set width, unit values
        and whether the wedge is British pica."""
        if not wedge_name:
            # use default S5
            return
        try:
            # For countries that use comma as decimal delimiter,
            # convert to point:
            wedge_name = str(wedge_name).replace(',', '.').upper().strip()
            # Check if this is an European wedge
            # (these were based on pica = .1667" )
            is_brit_pica = wedge_name.endswith('E')
            # Away with the initial S, final E and any spaces before and after
            # Make it work with space or dash as delimiter
            wedge = wedge_name.strip('SE ').replace('-', ' ').split(' ')
            series, raw_set_w = wedge
            # get the set width i.e. float approximated to .25
            set_w_str = ''.join(x for x in raw_set_w
                                if x.isnumeric() or x is '.')
            set_width = float(set_w_str) // 0.25 * 0.25
            # try to get the unit values, otherwise S5
            units = data.WEDGE_DEFINITIONS.get(series, d.S5)
            # update the attributes
            self.series, self.set_width = series, set_width
            self.is_brit_pica, self.units = is_brit_pica, units
        except (TypeError, AttributeError, ValueError):
            # In case parsing goes wrong
            raise ValueError('Cannot parse wedge name {}'.format(wedge_name))

    @property
    def parameters(self):
        """Gets a list of parameters"""
        em_inches = round(self.pica * self.set_width / 12, 5)
        parameters = OrderedDict()
        parameters['Wedge designation'] = self.name
        parameters['Inch width of 18 units [1em]'] = em_inches
        parameters['Units per row'] = ' '.join(str(x) for x in self.units if x)
        return parameters

    @property
    def units(self):
        """Gets the unit values for the wedge's rows"""
        units = self._units or d.S5
        # Add 0 at the beginning so that the list can be indexed with
        # row numbers - i.e. units[1] for row 1
        if units[0] is not 0:
            units = [0] + units
        # Ensure that wedge has 0 + 16 steps - i.e. length is 17
        # Keep adding the last value until we have enough
        while len(units) < 17:
            units.append(units[-1])
        return units

    @units.setter
    def units(self, units):
        """Sets the wedge unit values"""
        self._units = units

    def units_to_inches(self, units=1):
        """Convert units to inches, based on wedge's set width and pica def"""
        return round(units * self.set_width * self.pica / 216, 2)

    def get_adjustment_limits(self, low_space=False, cell_width=1):
        """Get the unit adjustment limits for this wedge.
        Without adjustment, the character has the same width as if the
        wedges were at 3/8.

        lower limit is reached at 1/1 0075 and 0005 wedge positions,
        max shrink is 3/8 - 1/1 = 2/7 => 2 * 0.0075 + 7 * 0.0005 = 0.0185"

        upper limit is reached at 15/15,
        max stretch is 15/15 - 3/8 = 12/7 => 12 * 0.0075 + 7 * 0.0005 = 0.0935"

        Constraints:
            Single-cell matrices have max .18" width for safety.
            Double-cell (large composition, wide characters) are .38"
            Otherwise the mould opening would not be covered by the matrix,
            leading to lead splashing over the diecase.

            Low spaces don't have this constraint as the upper mould blade
            prevents the lead from even reaching the diecase.

        Calculate this to wedge set units.
        Return positive integers.
        """
        # upper limit based on parameters
        stretch_inches = 0.0935
        cap = 0.18 + (cell_width - 1) * 0.2
        maximum = stretch_inches if low_space else min(stretch_inches, cap)
        # calculate adjustment range in units
        shrink = int(18 * 12 * 0.0185 / self.pica / self.set_width)
        stretch = int(18 * 12 * maximum / self.pica / self.set_width)
        return d.WedgeLimits(shrink, stretch)


class Layout:
    """Matrix case layout and outside characters data structure.
    layout : a tuple/list of tuples/lists denoting matrices:
              [(char, styles_string, position, units),...]
    diecase : a Diecase() class object

    Mats without position (None or empty string) will end up in the
    outside characters collection."""
    rows, columns = 15, 17
    used_mats, outside_mats = {}, []
    wedge = Wedge()

    def __repr__(self):
        return ('<Layout ({} rows, {} columns)>'
                .format(self.rows, self.columns))

    def __iter__(self):
        return (mat for mat in self.used_mats.values())

    def __contains__(self, obj):
        return obj in self.used_mats.values()

    @property
    def all_mats(self):
        """A list of all matrices - both used and outside"""
        return list(chain(self.used_mats.values(), self.outside_mats))

    @property
    def positions(self):
        """Gets all matrix positions for this layout"""
        by_row = product(self.row_numbers, self.column_numbers)
        return [d.Coordinates(column=col, row=row) for (row, col) in by_row]

    @property
    def row_numbers(self):
        """A tuple of row numbers"""
        return tuple(range(1, self.rows + 1))

    @property
    def column_numbers(self):
        """A tuple of column numbers."""
        if self.columns > 15 or self.rows > 15:
            return ['NI', 'NL', *'ABCDEFGHIJKLMNO']
        else:
            return [*'ABCDEFGHIJKLMNO']

    def select_row(self, row_number):
        """Get all matrices from a given row"""
        row_num = int(row_number)
        return [mat for (_, r), mat in self.used_mats.items() if r == row_num]

    def select_column(self, column):
        """Get all matrices from a given column"""
        col_num = column.upper()
        return [mat for (c, _), mat in self.used_mats.items() if c == col_num]

    def clean_layout(self):
        """Generate an empty layout of a given size."""
        # get coordinates to iterate over and build dict keys
        new_layout = OrderedDict().fromkeys(self.positions)
        # fill it with new empty matrices and define low/high spaces
        for position in new_layout:
            mat = bm.Matrix(char='', styles='*', diecase=self, units=0)
            mat.position = position
            # preset spaces
            mat.set_lowspace(position in d.DEFAULT_LOW_SPACE_POSITIONS)
            mat.set_highspace(position in d.DEFAULT_HIGH_SPACE_POSITIONS)
            new_layout[position] = mat
        return new_layout

    def get_charset(self, diecase_chars=True, outside_chars=False):
        """Diecase character set"""
        charset = {}
        used = self.used_mats.values() if diecase_chars else []
        unused = self.outside_mats if outside_chars else []
        for mat in chain(used, unused):
            if not mat.char or mat.isspace():
                continue
            for style in mat.styles:
                charset.setdefault(style, {})[mat.char] = mat
        return charset

    def get_lookup_table(self, diecase_chars=True, outside_chars=False):
        """Return a structure of {(mat char, style): mat}"""
        used = self.used_mats.values() if diecase_chars else []
        unused = self.outside_mats if outside_chars else []
        return {(mat.char, style): mat for mat in chain(used, unused)
                for style in mat.styles}

    def get_raw_layout(self):
        """Raw layout i.e. list of tuples with matrix parameters"""
        all_mats = chain(self.used_mats.values(), self.outside_mats)
        in_diecase = (mat.get_layout_record() for mat in all_mats)
        outside = (mat.get_layout_record(pos='') for mat in self.outside_mats)
        return [*in_diecase, *outside]

    def load_layout(self, source):
        """Parse an iterable or JSON-encoded array to read a layout.
        Sort the layout into matrices in diecase and outside chars.
        Accepts any iterator of (char, style_string, position, units)"""
        try:
            raw_data = json.loads(source)
        except (TypeError, json.JSONDecodeError):
            raw_data = source

        try:
            # Get matrices from supplied layout's canonical form
            recs = (d.MatrixRecord(*record) for record in raw_data)
            mats = [bm.Matrix(char=rec.char, styles=rec.styles, code=rec.code,
                              units=rec.units, diecase=self) for rec in recs]
            # parse the source layout to get its size,
            # reversing the order increases the chance of finding row 16 faster
            for matrix in reversed(mats):
                if matrix.position.row == 16:
                    # finish here as the only 16-row diecases were 16x17
                    self.rows, self.columns = 16, 17
                    break
                if matrix.position.column in ('NI', 'NL'):
                    # update the columns number
                    # iterate further because we can still find 16th row
                    self.columns = 17
        except (TypeError, ValueError):
            # Layout supplied is incorrect; use a default size of 15x17
            mats, self.rows, self.columns = [], 15, 17
        # Build empty layout determining its size
        used_mats = self.clean_layout()
        # Fill it with matrices for existing positions
        for mat in mats[:]:
            position = mat.position
            if used_mats.get(position):
                used_mats[position] = mat
                mats.remove(mat)
        # The remaining mats will be stored in outside layout
        self.used_mats, self.outside_mats = used_mats, mats

    def purge(self):
        """Resets the layout to an empty one of the same size"""
        self.outside_mats = []
        self.used_mats = self.clean_layout()

    def resize(self, rows=15, columns=17):
        """Rebuild the layout to adjust it to the new diecase format"""
        # manipulate data structures locally
        old_layout, old_extras = self.used_mats, self.outside_mats
        new_extras = []
        # resize
        self.rows, self.columns = rows, columns
        new_layout = self.clean_layout()
        # preserve mats as outside characters when downsizing the layout
        for position, mat in old_layout.items():
            if new_layout.get(position):
                # the new layout has a position for this matrix
                new_layout[position] = mat
            else:
                # no place for it = put it in outside chars
                new_extras.append(mat)
        # pull the mats from outside layout to diecase automatically
        # if so, remove them from outside layout
        for mat in old_extras[:]:
            position = mat.position
            if new_layout.get(position) is not None:
                # there is something at this position
                new_layout[position] = mat
                old_extras.remove(mat)
        # finally update the instance attributes
        self.used_mats = new_layout
        self.outside_mats = old_extras + new_extras

    def select_many(self, char=None, styles=None, code=None, units=None,
                    islowspace=None, ishighspace=None):
        """Get all matrices with matching parameters"""
        def match(matrix):
            """Test the matrix for conditions unless they evaluate to False"""
            # guard against returning True if no conditions are valid
            # end right away
            wants = (char, styles, code, units, islowspace, ishighspace)
            if not any(wants):
                return False
            finds = (matrix.char, matrix.styles, matrix.code, matrix.units,
                     matrix.islowspace(), matrix.ishighspace())
            # check all that apply and break on first mismatch
            for wanted, found in zip(wants, finds):
                if wanted and wanted != found:
                    return False
            # all conditions match
            return True

        results = [mat for mat in self.used_mats.values() if match(mat)]
        return results

    def select_one(self, *args, **kwargs):
        """Get a matrix with denoted parameters"""
        with suppress(IndexError):
            search_results = self.select_many(*args, **kwargs)
            return search_results[0]
        raise bm.MatrixNotFound

    @property
    def spaces(self):
        """Get all available spaces"""
        return [mat for mat in self.used_mats.values() if mat.isspace()]

    def get_space(self, units=0, low=True, wedge=None):
        """Find a suitable space in the diecase layout"""
        def mismatch(checked_space):
            """Calculate the unit difference between space's width
            and desired unit width"""
            wdg = wedge if wedge else self.wedge
            # how much adjustment would be needed? single or double mat?
            difference = units - wdg[checked_space.position.row]
            low = checked_space.islowspace()
            cells = checked_space.size.horizontal
            # calculate minimum and maximum
            limits = wdg.get_adjustment_limits(low_space=low, cell_width=cells)
            shrink, stretch = limits.shrink, limits.stretch
            return abs(difference) if -shrink < difference < stretch else -1

        spaces = [mat for mat in self.spaces if mismatch(mat) >= 0 and
                  mat.islowspace() == low and mat.ishighspace() != low]
        matches = sorted(spaces, key=mismatch)
        try:
            return matches[0]
        except IndexError:
            # can't match a space no matter how we try
            exc_message = 'Cannot find a {} space close enough to {} units.'
            high_or_low = 'low' if low else 'high'
            raise bm.MatrixNotFound(exc_message.format(high_or_low, units))

    def get_min_space_width(self, high_space=False):
        """Find the minimum space width for the diecase"""
        return min(space.get_min_units() for space in self.spaces
                   if space.ishighspace() == high_space)

    def get_max_space_width(self, high_space=False):
        """Find the maximum space width for the diecase"""
        return max(space.get_max_units() for space in self.spaces
                   if space.ishighspace() == high_space)


class Diecase(Layout, BaseModel):
    """Diecase: matrix case attributes and operations"""
    diecase_id = pw.TextField(db_column='diecase_id', primary_key=True)
    _typeface = pw.TextField(db_column='typeface', null=True,
                             help_text='JSON-encoded typeface metadata')
    _wedge = pw.TextField(db_column='wedge_name', default='S5-12E',
                          help_text='wedge series and set width')
    _layout = pw.TextField(db_column='layout',
                           help_text='JSON-encoded diecase layout')

    def __repr__(self):
        return ('<Diecase: {} - {}>'
                .format(self.diecase_id, self.description))

    def __str__(self):
        return self.diecase_id

    def __bool__(self):
        return bool(self.diecase_id)

    def __call__(self):
        return self

    @property
    def description(self):
        """Get the diecase description"""
        if not self.typeface_data:
            return self._typeface
        faces = [face for (face, _) in self.typeface_data.values()]
        series = '+'.join(sorted({tf.series for tf in faces}))
        names = ', '.join(sorted({tf.name for tf in faces}))
        return '{} - {}'.format(series, names)

    @property
    def parameters(self):
        """Gets a list of parameters"""
        parameters = OrderedDict()
        parameters['Diecase ID'] = self.diecase_id
        parameters['Typeface'] = self.description
        parameters['Assigned wedge'] = self.wedge.name
        return parameters

    @property
    def layout_styles(self):
        """Get all available character styles from the diecase layout."""
        return sum(mat.styles for mat in self.all_mats
                   if not mat.styles.use_all) or bm.Styles(None)

    @property
    def typeface_styles(self):
        """Get the typeface styles for this diecase"""
        return bm.Styles(self.typeface_data)

    @property
    def styles(self):
        """Get all styles, both defined for typeface and for layout"""
        return self.typeface_styles + self.layout_styles

    @property
    def typeface_data(self):
        """Typeface data"""
        _cached = self.__dict__.get('_typeface_data')
        if _cached:
            # early return = don't calculate it again
            return _cached

        try:
            source = json.loads(self._typeface)
        except (json.JSONDecodeError, TypeError):
            # revert to default 327-12
            typeface = TypefaceVariant()
            typeface_tuple = (typeface, typeface.unit_arrangement)
            _typeface_data = OrderedDict({typeface.style: typeface_tuple})
            self.__dict__['_typeface_data'] = _typeface_data
            return _typeface_data

        # each style (roman, bold, italic) has a typeface and UA assigned
        # e.g. {'r': ('327', '12D', 'r', '325', 'r'),
        #       'i': ('327', '12D', 'i', '324', 'r')}
        _typeface_data = OrderedDict()
        for style in bm.Styles(source.keys()):
            type_id, size, style_id, ua_id, ua_variant = source[style.short]
            typeface_variant = TypefaceVariant(type_id, size, style_id)
            try:
                unit_arrangement = UAVariant(ua_id, ua_variant)
            except ValueError:
                try:
                    unit_arrangement = UAVariant(ua_id, style.short)
                except ValueError:
                    unit_arrangement = typeface_variant.unit_arrangement
            # update the data
            _typeface_data[style] = (typeface_variant, unit_arrangement)

        # data ready to ship (cache to avoid recalculating)
        self.__dict__['_typeface_data'] = _typeface_data
        return _typeface_data

    @typeface_data.setter
    def typeface_data(self, typeface_data):
        """Set the typeface metadata"""
        # prepare the output data
        output = OrderedDict()
        for style, (tface, arrangement) in typeface_data.items():
            output[style.short] = (tface.series, tface.size, tface.style.short,
                                   arrangement.number, arrangement.short)
        self._typeface = json.dumps(output)
        self.__dict__['_typeface_data'] = typeface_data

    @property
    def unit_arrangements(self):
        """Return a mapping of UAVariant objects to styles."""
        return {s: ua for s, (_, ua) in self.typeface_data.items()}

    def load(self):
        """Loads all data - required when creating a new instance,
        or reading one from database"""
        self.load_layout(self._layout)
        self.wedge = Wedge(wedge_name=self._wedge)
        return self

    def store(self):
        """Store information for saving in database"""
        self.store_layout()
        self._wedge = self.wedge.name

    def store_layout(self):
        """Save the layout canonical form to ORM"""
        self._layout = json.dumps(self.get_raw_layout())


class UnitArrangement:
    """Unit arrangement definition.
    Can contain one or more variants for different type styles.
    """
    variant_sequence = ('r', 'i', 's', 'ar', 'ai', 'as', 'br', 'g',
                        's1', 's2', 's3', 's4', 's5')
    variant_definitions = dict(r=('regular', 'roman or the only style'),
                               i=('italic', ''),
                               s=('small caps', ''),
                               g=('Gaelic accents', ''),
                               ar=('regular A', 'modified char widths'),
                               ai=('italic A', 'modified char widths'),
                               br=('regular B', 'modified char widths'),
                               s1=('size 1', 'titling'),
                               s2=('size 2', 'titling'),
                               s3=('size 3', 'titling'),
                               s4=('size 4', 'titling'),
                               s5=('size 5', 'titling'))
    variant_definitions['as'] = ('small caps A', 'modified char widths')

    def __init__(self, number=0):
        # store the number (string in case int was supplied)
        self.number = str(number)

        # look up the unit arrangement data
        arrangement_data = data.UNIT_ARRANGEMENTS.get(self.number, {})
        if not arrangement_data:
            raise ValueError('Unit arrangement {} not found!'.format(number))

        # store the underlying data
        self._raw_data = OrderedDict([(variant, arrangement_data[variant])
                                      for variant in self.variant_sequence
                                      if arrangement_data.get(variant)])

    def __str__(self):
        return 'UA #{}'.format(self.number)

    def __repr__(self):
        return '<UnitArrangement #{}>'.format(self.number)

    def __getitem__(self, item):
        return self(item)

    def __call__(self, variant):
        return self.get_variant(variant)

    def get_variant(self, variant):
        """Make a UAVariant instance for a specified variant"""
        return UAVariant(self.number, variant)

    @property
    def variants(self):
        """Get all variants for this unit arrangement"""
        return [self.get_variant(var) for var in self._raw_data]

    @property
    def variant_names(self):
        """Get a string with UA variant names and letters"""
        return ',\n'.join('{}: {}'.format(s, self.variant_definitions[s][0])
                          for s in self.variant_sequence)


class UAVariant(UnitArrangement):
    """Unit arrangement variant class. Stores unit values for characters.
    Each unit arrangement has one or more variants
    (for example: roman, italic, small caps, Gaelic accents,
    size 1, 2, 3, 4, 5 - mostly for titling,
    modifications with different units for some characters)."""
    def __init__(self, number, variant='r'):
        super().__init__(number)
        self.short = variant

        # is this the valid unit arrangement variant?
        variant_definition = self.variant_definitions.get(variant)
        if not variant_definition:
            raise ValueError('Incorrect variant: {}'.format(variant))
        self.name, self.alternatives = variant_definition

        # get the unit values for this variant
        arrangement = self._raw_data.get(self.short, {})
        if not arrangement:
            raise ValueError('{} has no character unit values!'.format(self))

        # parse the UA
        by_units = defaultdict(list)
        for char, units in sorted(arrangement.items()):
            by_units[units].append(char)
        # store the result
        self.by_char, self.by_units = arrangement, by_units

    def __str__(self):
        return 'Unit arrangement #{} {}'.format(self.number, self.name)

    def __repr__(self):
        return '<UAVariant: #{} {}>'.format(self.number, self.name)

    def __call__(self, item):
        """Get a unit arrangement for a given style"""
        @singledispatch
        def getter(_):
            """dispatch on type, accept only strings or numeric values"""
            raise TypeError('Unknown type')

        getter.register(int, self.get_chars)
        getter.register(float, self.get_chars)
        getter.register(str, self.get_units)
        return getter(item)

    def get_chars(self, unit_value):
        """Get the characters list for a given unit value"""
        value = int(unit_value)
        return sorted(self.by_units.get(value))

    def get_units(self, char):
        """Get unit value from an arrangement"""
        # first try to look it up in arrangement
        with suppress(KeyError):
            return self.by_char[char]
        # might be an accented version
        base_char = d.REVERSE_ACCENTS.get(char)
        if not char:
            return 0
        with suppress(KeyError):
            return self.by_char[base_char]
        # no character in UA
        raise ValueError('Unit value for {} not found!'.format(char))

    def get_mat_units(self, matrix):
        """Get unit value for a Matrix object"""
        return self.get_units(matrix.char)


class Typeface:
    """Typeface definition. Gets all data for the given typeface number.

    Typeface data is stored in data/typefaces.json as a dictionary:
        {"262": {"name": "Gill Sans",
                 "scripts": ["Latin", "Cyrillic"],
                 "tags": ["sans serif", "modern", "humanist", "Eric Gill"],
                 "related": {"b": (275, "r"), "q": (275, "i")},
                 "sizes": {"6": [6.5, .1220],
                           "8": [8.25, .1285] etc.},
                 "uas": {"r": (82, "r"), "i": (82, "i")}},
        "327": {"name": "Times New Roman"...}}
    Keys in the main dictionary are typeface series numbers.
    Values are the typeface data:
        name - typeface name,
        scripts - a list of scripts/alphabets the typeface was designed for,
        tags - a list of tags describing the typeface (e.g. Vox-ATypI classes),
        related - a dictionary of related typefaces by style,
        sizes - size-specific data: [set_width, calibration_line, ua_overrides]
            where calibration_line is the "M line" parameter from the
            type specimen book, and ua_overrides are optional corrections
            to the unit arrangement mappings for some sizes or styles,
        uas - unit arrangement mapping for each of the styles.
    """
    def __init__(self, series='327'):
        def get_size_data(size):
            """Get the size data from typeface data"""
            size_data = raw_sizes.get(size)
            with suppress(TypeError, AttributeError):
                # first try getting a reference to another size
                # it's a string so it should be convertible to uppercase
                return get_size_data(size_data.upper().strip())
            # check if we have a tuple/list of 3 or 2 items
            with suppress(TypeError, ValueError):
                # 3 values: set width, M-line, unit arrangement overrides
                set_width, calibration_line, raw_ua_data = size_data
                return set_width, calibration_line, raw_ua_data
            with suppress(TypeError, ValueError):
                # check if it's a 2-item tuple or list
                set_width, calibration_line = size_data
                return set_width, calibration_line, {}
            # nothing above is correct? (for example, lookup returned None)
            return 0, 0, {}

        def sort_key(size):
            """Sorting key for sizes: get the number at the front"""
            chars = []
            for char in size:
                if char.isdigit():
                    chars.append(char)
                else:
                    break
            return int(''.join(chars) or '0')

        self.series = str(series)
        raw_data = data.TYPEFACES.get(self.series, {})
        self._raw_data = raw_data
        self.name = raw_data.get('name', 'unknown typeface')
        self.classification = ', '.join(raw_data.get('tags', ['']))
        self.script = ', '.join(raw_data.get('scripts', ['Latin']))
        uas = raw_data.get('uas', {})
        raw_sizes = self._raw_data.get('sizes', {})
        self.sizes = OrderedDict((size, get_size_data(size))
                                 for size in sorted(raw_sizes, key=sort_key))
        self.unit_arrangements = {bm.STYLES.get(st): UAVariant(ua_id, ua_v)
                                  for st, (ua_id, ua_v) in uas.items()}

    def __call__(self, size):
        return self.get_size(size)

    def __getitem__(self, item):
        return self(item)

    def __repr__(self):
        return '<Typeface: {t.series} {t.name}>'.format(t=self)

    def __str__(self):
        return ('{t.series} {t.name} ({t.script} {stylenames})'
                .format(t=self, stylenames=self.styles.names))

    def get_size(self, size):
        """Get a variant in a specified size"""
        return TypefaceSize(self.series, size)

    @property
    def related(self):
        """Get related (combinable in diecase) typefaces for this typeface,
        e.g. 334 Times Bold for 327 Times New Roman"""
        related = self._raw_data.get('related', {})
        roman = bm.Styles.definitions.roman
        retval = {}
        # related styles are defined similar to unit arrangements:
        # {rel_style: (typeface_number, style_in_typeface)}
        for raw_style, (rel_tface, rel_style) in related.items():
            style = bm.STYLES.get(raw_style)
            if not style:
                continue
            related_style = bm.STYLES.get(rel_style) or roman
            tface = Typeface(rel_tface)
            retval[style] = (tface, related_style)

        return retval

    @property
    def styles(self):
        """Get the styles for this typeface."""
        # DEPRECATED: styles will be stored in unit arrangement mappigns only!
        # use the "styles" as a fallback only
        typeface_styles = self._raw_data.get('styles', '')
        return bm.Styles(self.unit_arrangements or typeface_styles)

    @property
    def related_styles(self):
        """Get all related styles. These are combinable in the same diecase."""
        return bm.Styles(self.related)

    @property
    def combined_styles(self):
        """Get all styles for this typeface and related typefaces combined."""
        return self.styles + self.related_styles


class TypefaceSize(Typeface):
    """A typeface in a specified size."""
    def __init__(self, series='327', size='12D'):
        super().__init__(series)
        converted_size = str(size).upper().strip()
        # strip all non-numeric characters
        if converted_size not in self.sizes:
            converted_size = ''.join(x for x in converted_size if x.isdigit())
        self.size = converted_size
        set_width, m_line, raw_ua_data = self.sizes.get(self.size, (0, 0, {}))
        # unit arrangement override (mostly for larger sizes)
        ua_overrides = {bm.STYLES.get(st): UAVariant(ua_id, ua_v)
                        for st, (ua_id, ua_v) in raw_ua_data.items()}
        self.unit_arrangements.update(ua_overrides)
        self.set_width, self.calibration_line = set_width, m_line

    def __repr__(self):
        return '<TypefaceSize: {t.series}-{t.size} {t.name}>'.format(t=self)

    def __str__(self):
        return ('{t.series}-{t.size} {t.name} ({t.script} {stylenames})'
                .format(stylenames=self.styles.names, t=self))

    def __call__(self, style):
        return self.get_variant(style)

    @property
    def related(self):
        """Get related (combinable in diecase) TypefaceSize objects,
        e.g. 334 Times Bold for 327 Times New Roman"""
        related = super().related
        return {style: tface(self.size)(rel_style)
                for style, (tface, rel_style) in related.items()}

    def get_variant(self, style):
        """Get a typeface variant for a given style"""
        _style = bm.STYLES.get(style)
        if not _style:
            raise ValueError('Incorrect style: {}'.format(style))
        try:
            return TypefaceVariant(self.series, self.size, _style)
        except ValueError:
            # look it up in related typefaces
            with suppress(KeyError):
                return self.related[_style]
            msg = ('{} not present in {}-{} {} or any of its related typefaces'
                   .format(_style.name, self.series, self.size, self.name))
            raise ValueError(msg)


class TypefaceVariant(TypefaceSize):
    """A concrete typeface/size/style combination."""
    def __init__(self, series='327', size='12D', style=None):
        super().__init__(series, size)
        _style = (bm.STYLES.get(style) or
                  (self.styles.first if style is None else None))
        if not _style:
            raise ValueError('Incorrect style: {}'.format(style))
        elif _style not in self.styles:
            msg = ('{} not present in {}-{} {}'
                   .format(_style.name, self.series, self.size, self.name))
            raise ValueError(msg)
        self.style = _style
        self.unit_arrangement = self.unit_arrangements.get(_style, {})

    def __repr__(self):
        stylename = self.style.name if self.style else self.style
        return ('<TypefaceVariant: {t.series}-{t.size} {t.name} {style}>'
                .format(t=self, style=stylename))

    def __str__(self):
        return '{t.series}-{t.size} {t.name} {t.style.name}'.format(t=self)


class Ribbon(BaseModel):
    """Ribbon objects - no matter whether files or database entries.

    A ribbon has the following attributes:
    description, customer - strings,
    diecase_id (diecase is selected automatically on casting, so user can e.g.
                view the layout or know which wedge to use),
    source_text - original text the ribbon was set from; useful for
                  re-composing for different parameters,
    contents - series of Monotype codes.

    Methods:
    choose_ribbon - choose ribbon automatically or manually,
        first try to get a ribbon with ribbon_id, and if that fails
        ask and select ribbon manually from database, and if that fails
        ask and load ribbon from file
    read_from_file - select a file, parse the metadata, set the attributes
    export_to_file - store the metadata and contents in text file
    store_in_db - store the metadata and contents in db
    set_[description, customer, diecase_id] - set parameters manually"""
    ribbon_id = pw.TextField(db_column='ribbon_id', primary_key=True,
                             default='New Ribbon',
                             help_text='Unique ribbon name')
    description = pw.TextField(db_column='description', default='')
    customer = pw.TextField(db_column='customer', default='')
    diecase = pw.ForeignKeyField(Diecase)
    wedge_name = pw.TextField(db_column='wedge_name', default='')
    source_text = pw.TextField(db_column='source_text', default='')
    contents = pw.TextField(db_column='contents', default='')

    def __iter__(self):
        return iter(self.contents)

    def __next__(self):
        yield from self.contents

    def __repr__(self):
        return self.ribbon_id or ''

    def __bool__(self):
        return bool(self.contents)

    def __call__(self):
        return self

    @property
    def parameters(self):
        """Gets a list of parameters"""
        parameters = OrderedDict()
        parameters['Ribbon ID'] = self.ribbon_id
        parameters['Description'] = self.description
        parameters['Customer'] = self.customer
        parameters['Diecase ID'] = self.diecase_id
        parameters['Wedge'] = self.wedge_name
        return parameters

    def update(self, source=None):
        """Updates the object attributes with a dictionary"""
        # Allow to use this method to initialize a new empty ribbon
        if not source:
            source = {}
        with suppress(AttributeError):
            self.ribbon_id = source.get('ribbon_id', '')
            self.description = source.get('description', '')
            self.customer = source.get('customer', '')
            self.diecase_id = source.get('diecase_id', '')
            self.wedge_name = source.get('wedge_name', '')
            self.contents = source.get('contents', [])

    def import_from_file(self, ribbon_file):
        """Imports ribbon from file, parses parameters, sets attributes"""
        with suppress(AttributeError):
            # Try to open it and get only the lines containing non-whitespace
            with ribbon_file:
                raw_data = (line.strip() for line in ribbon_file.readlines())
                ribbon = [line for line in raw_data if line]
            metadata = p.parse_ribbon(ribbon)
            # Update the attributes with what we found
            self.update(metadata)

    def export_to_file(self, ribbon_file):
        """Exports the ribbon to a text file"""
        # Choose file, write metadata, write contents
        with ribbon_file:
            ribbon_file.write('description: ' + self.description)
            ribbon_file.write('customer: ' + self.customer)
            ribbon_file.write('diecase: ' + self.diecase_id)
            ribbon_file.write('wedge: ' + self.wedge_name)
            for line in self.contents:
                ribbon_file.write(line)



# -*- coding: utf-8 -*-
"""User interface for rpi2caster. Text UI is implemented here, additional UIs
can be added later or imported from separate modules"""
import readline
import glob
import string
from collections import namedtuple, OrderedDict
from contextlib import suppress
from functools import partial
import click
from . import datatypes as dt


# menu item namedtuple
MenuItem = namedtuple('Item', 'key value condition text desc seq')
# control key definitions
Key = namedtuple('Key', 'getchar name')
KEYS = dict(
    # main keys
    esc=Key('\x1b', 'Esc'), enter=Key('\r', 'Enter'), space=Key(' ', 'Space'),
    tab=Key('\t', 'Tab'), backspace=Key('\x7f', 'Backspace'),
    # arrow keys
    up=Key('\x1b[A', 'Up'), down=Key('\x1b[B', 'Down'),
    left=Key('\x1b[D', 'Left'), right=Key('\x1b[C', 'Right'),
    # editor keys
    ins=Key('\x1b[2~', 'Ins'), delete=Key('\x1b[3~', 'Del'),
    home=Key('\x1b[H', 'Home'), end=Key('\x1b[F', 'End'),
    pgup=Key('\x1b[5~', 'PgUp'), pgdn=Key('\x1b[6~', 'PgDn'),
    # ctrl combinations
    ctrl_c=Key('\x03', 'Ctrl-C'), ctrl_z=Key('\x1a', 'Ctrl-Z'),
    ctrl_s=Key('\x13', 'Ctrl-S'), ctrl_n=Key('\x0e', 'Ctrl-N'),
    # function keys
    f1=Key('\x1bOP', 'F1'), f2=Key('\x1bOQ', 'F2'),
    f3=Key('\x1bOR', 'F3'), f4=Key('\x1bOS', 'F4'),
    f5=Key('\x1b[15~', 'F5'), f6=Key('\x1b[17~', 'F6'),
    f7=Key('\x1b[18~', 'F7'), f8=Key('\x1b[19~', 'F8'),
    f9=Key('\x1b[20~', 'F9'), f10=Key('\x1b[21~', 'F10'),
    f11=Key('\x1b[23~', 'F11'), f12=Key('\x1b[24~', 'F12')
    )
DEFAULT_ABORT_KEYS = [KEYS[key] for key in ('esc', 'f10')]
NONSENSE = 'etaoin shrdlu cmfwyp'


def get_key(source):
    """Get the Key namedtuple of a key. First look it up in special keys."""
    # normalize to lowercase and strip all whitespace ('Esc' -> 'esc')
    # replace spaces and dashes to underscores ('ctrl-C -> ctrl_c)
    if source is None:
        # can be used to generate numbers on the fly
        return Key(getchar=None, name=None)
    else:
        key = str(source)
        normalized_key = key.lower().strip()
        normalized_key.replace('-', '_')
        normalized_key.replace(' ', '_')
        return KEYS.get(normalized_key) or Key(getchar=key, name=key)


def option(key=None, value=None, cond=True, text='', desc='', seq=50):
    """Menu item factory for UI menus:
        key   : key to press to choose that option (int, str or None),
                    if None, key will be autogenerated,
        value : value to be returned; will be raised if it is an exception
        cond  : condition to check (function or value),
                    an option is available if this evaluates to True,
        text  : short description of an option,
                    optional (allows "hidden" options too)
        desc  : additional help/description,
        seq   : menu item sequence number for arbitrary sorting.
    """
    return MenuItem(get_key(key), value, cond, str(text), str(desc), seq)


def assess(condition, force_bool=False):
    """Try to call the condition and evaluate the retval,
    and if it fails, evaluate the conditional itself.

    Allows lazily evaluating functions, useful for dynamic menus.

    force_bool : converts value to boolean using Python rules."""
    try:
        retval = condition()
    except TypeError:
        # check is not callable
        retval = condition
    return bool(retval) if force_bool else retval


def get_sorted_valid_options(options):
    """Iterate over a list of options and return a dictionary of options
    which will be displayed in menu.

    Any option with unmet condition will be omitted.

    Options will be sorted first by option.seq parameter
    (allows arbitrary menu item positioning), then by displayed key,
    and finally, options with unspecified keys will be assigned
    to numbers, then lowercase and uppercase letters.
    """
    def sort_function(opt):
        """Returns a tuple for sorting options"""
        # sort by sequence, then lowercase key name, then text and description
        # missing parameters will be placed at the bottom
        return (opt.seq,
                opt.key.name.lower() if opt.key.name else 'zzz',
                opt.text if opt.text else 'zzz',
                opt.desc if opt.desc else 'zzz')

    # input data format: (MenuItem1, MenuItem2...)
    # first check if option condition is met
    valid_opts = [option for option in assess(options)
                  if assess(option.condition, force_bool=True)]
    # generate the sequence of keys from 1 to 9, lowercase, uppercase
    # reject all keys already assigned to options
    used_keys = [option.key.name for option in valid_opts if option.key.name]
    autokeys = (string.digits[1:] + string.ascii_letters)
    key_generator = (get_key(k) for k in autokeys if k not in used_keys)
    # ...and apply them to options with missing keys
    ret = [option if (option.key.getchar and option.key.name)
           else option._replace(key=next(key_generator))
           for option in sorted(valid_opts, key=sort_function)]
    # finally, sort the options by their keys and return
    sorted_ret = sorted(ret, key=sort_function)
    return sorted_ret


def build_entry(opt, trailing_newline=1):
    """Build a menu entry for an option"""
    # templates for menu entries
    long_entry = '\t{key}:  {text}\n\t\t\t{desc}' + '\n' * trailing_newline
    short_entry = '\t{key}:  {text}' + '\n' * trailing_newline
    key_name = click.style('{:<10}'.format(opt.key.name), bold=True, fg='cyan')
    # use a proper template depending on available data
    if opt.desc:
        return long_entry.format(key=key_name, text=opt.text, desc=opt.desc)
    elif opt.text:
        return short_entry.format(key=key_name, text=opt.text)


def build_menu_prompt(menu_actions, default_retval, default_key, abort_keys):
    """Make a menu prompt displaying default option and exit keys."""
    enter_key = get_key('enter')
    # add default key combo if it was specified
    if menu_actions.get(enter_key.getchar) == default_retval != NONSENSE:
        chunks = [click.style('[Enter]', fg='cyan', bold=True),
                  click.style(' chooses option for key [', fg='cyan'),
                  click.style(default_key.name, fg='cyan', bold=True),
                  click.style('].', fg='cyan')]
        def_text = ''.join(chunks)
    else:
        def_text = ''

    # which keys abort the menu?
    keys = sorted(key.name for key in abort_keys)
    if keys:
        # additional info if abort is allowed
        chunks = [click.style('Press ', fg='cyan'),
                  click.style('[', fg='cyan', bold=True),
                  click.style(', '.join(keys), fg='cyan', bold=True),
                  click.style(']', fg='cyan', bold=True),
                  click.style(' to exit.', fg='cyan')]
        abort_s = ''.join(chunks)
    else:
        abort_s = ''

    prompt = click.style('\nYour choice? :', fg='yellow', bold=True)
    return ' '.join(t for t in [def_text, abort_s, prompt] if t)


class Abort(click.Abort):
    """Exception - abort the current action"""
    def __str__(self):
        return ''


class Finish(click.Abort):
    """Exception - finish the current exit to menu etc."""
    def __str__(self):
        return ''


class ClickUI(object):
    """Click-based text user interface"""
    __name__ = 'Text UI based on Click'

    def __init__(self, verbosity=0):
        self.verbosity = verbosity

    def dynamic_menu(self, options,
                     header='', footer='',
                     default_key=None,
                     allow_abort=True,
                     func_args=None, func_kwargs=None,
                     catch_exceptions=(Abort, click.Abort,
                                       KeyboardInterrupt, EOFError)):
        """dynamic_menu(options=[MenuItem1, MenuItem2...],
                        header='', footer='',
                        default_key=None, allow_abort=True,
                        func_args=None, func_kwargs=None,
                        catch_exceptions=[exceptions]):

        Builds menu dynamically and executes the callback with no arguments
        or raises an exception each time.
        After callback is executed, builds menu again and displays
        refreshed options.
        Raising an exception here allows to return from menu.

        options : option definitions iterable: (MenuItem1, MenuItem2...),

                  each MenuItem has following attributes/fields:

                  key (int or str) : key to press to choose that option,

                  value : function or exception to be called or raised;
                          function will be called with
                          *func_args and **func_kwargs

                  cond : condition to check (function or value),
                         an option is available if this evaluates to True,

                  text : short description of an option,
                         optional (allows "hidden" options too),

                  desc : additional help/description,

                  seq : menu item sequence number for arbitrary sorting.

        header, footer : will be displayed above and below menu options list,

        default_key : if defined and valid, pressing Enter will choose
                      an option defined for this key,

        allow_abort : lets Esc or Ctrl-C raise Abort,

        func_args=(), func_kwargs={} : positional and keyword arguments
                                       passed to the function call,

        catch_exceptions : an iterable of exceptions this menu is supposed
                           to catch, rather than bubbling up and crashing
                           the program
        """
        # no args or kwargs? use empty tuple or dict then
        args = func_args or ()
        kwargs = func_kwargs or {}
        # keep displaying the menu until exception is raised
        while True:
            ret = self.menu(options, header, footer, default_key, allow_abort)
            # if option value is an expression, it'll bubble up here
            with suppress(catch_exceptions):
                ret(*args, **kwargs)

    def menu(self, options, header='', footer='',
             default_key=None, allow_abort=True):
        """menu(options=[MenuOption1, MenuOption2,...]
                header='', footer='', default_key=None, allow_abort=True):

        A menu displaying keys, options and descriptions (if provided).

        User is asked to press a key and then the function tries to
        raise an exception (if option.value is Exception subclass)
        or return option.value.

        Additional behavior can be specified with default_key and
        allow_abort options in the function call.

        options : option definitions iterable: (MenuItem1, MenuItem2...),

                  each MenuItem has following attributes/fields:

                  key (int or str) : key to press to choose that option,

                  value : value to be returned;
                          will be raised if it is an exception

                  cond : condition to check (function or value),
                         an option is available if this evaluates to True,

                  text : short description of an option,
                         optional (allows "hidden" options too)

                  desc : additional help/description,

                  seq : menu item sequence number for arbitrary sorting.

        header : string to be displayed above option entries,

        footer : string to be displayed below option entries,

        default_key : if Enter is pressed, option for this key
                      will be selected instead (provided that option
                      is defined, valid and condition is met),

        allow_abort : raise Abort if Esc or Ctrl-C is pressed,
                      otherwise stay in the menu. If any of these keycombos
                      is used as a key for a menu option,
                      it will select that option rather than aborting.

        After choice is made, try to raise the corresponding option
        or return it.
        """
        def build_menu():
            """Build a list of options and their descriptions"""
            # header_string: displayed above, footer_string: displayed below
            # debug_string: displayed only if verbosity != 0
            # accept functions for dynamic content generation as well
            _header = assess(header)
            _footer = assess(footer)
            header_text = '\n{}\n'.format(_header) if _header else ''
            footer_text = '\n{}\n'.format(_footer) if _footer else '\n'
            debug = ('\nThe program is now in debugging mode. Verbosity: {}\n'
                     .format(self.verbosity) if self.verbosity else '')

            # generate menu entries for options
            entries = (build_entry(o) for o in valid_options)
            # return the newly constructed menu list
            return [header_text, debug, *entries, footer_text]

        # generate a list of valid options
        valid_options = get_sorted_valid_options(options)
        # dictionary of key: return value pairs
        rets = {option.key.getchar: option.value
                for option in valid_options}
        # get default return value for wrong keypress if we have default_key
        # default to some nonsensical string for wrong dict hits
        # this will make it possible to have None as an option
        def_key, enter_key = get_key(default_key), get_key('enter')
        def_rv = rets.get(def_key.getchar, NONSENSE)
        if enter_key.getchar not in rets:
            rets[enter_key.getchar] = def_rv
        # check which keys can be used for aborting the menu
        # pressing one of them would raise Abort, if two conditions are met:
        # aborting is enabled (allow_abort=True) and key is not in options
        abort_keys = [key for key in DEFAULT_ABORT_KEYS
                      if allow_abort and key.getchar not in rets]
        # display the menu
        # clear the screen, display the header, options, footer
        click.clear()
        click.echo('\n'.join(build_menu()))
        click.echo(build_menu_prompt(rets, def_rv, def_key, abort_keys))
        # wait for user input
        while True:
            getchar = click.getchar()
            if getchar in [key.getchar for key in abort_keys]:
                raise Abort
            retval = rets.get(getchar, NONSENSE)
            if retval != NONSENSE:
                return dt.try_raising(retval)

    @staticmethod
    def simple_menu(message, options, default_key=None, allow_abort=True):
        """A simple menu where user is asked what to do.
        Wrong choice points back to the menu if default_option is not defined.

        message : string displayed on screen,

        options : a list of MenuItem namedtuples,

        default_key : default key if Enter is pressed,

        allow_abort : allow aborting with ctrl-C or Esc.
        """
        # generate a list of valid options
        valid_options = get_sorted_valid_options(options)
        entries = (build_entry(option, trailing_newline=0)
                   for option in valid_options)
        # dictionary of key: return value pairs
        rets = {option.key.getchar: option.value
                for option in valid_options}
        # get default return value for wrong keypress if we have default_key
        # default to some nonsensical string for wrong dict hits
        # this will make it possible to have None as an option
        def_key, enter_key = get_key(default_key), get_key('enter')
        def_rv = rets.get(def_key.getchar, NONSENSE)
        if enter_key.getchar not in rets:
            rets[enter_key.getchar] = def_rv
        # check which keys can be used for aborting the menu
        # pressing one of them would raise Abort, if two conditions are met:
        # aborting is enabled (allow_abort=True) and key is not in options
        abort_keys = [key for key in DEFAULT_ABORT_KEYS
                      if allow_abort and key.getchar not in rets]
        # display the menu
        click.echo('\n'.join(['', assess(message), '', *entries, '']))
        click.echo(build_menu_prompt(rets, def_rv, def_key, abort_keys))
        # Wait for user input
        while True:
            getchar = click.getchar()
            if getchar in [key.getchar for key in abort_keys]:
                raise Abort
            retval = rets.get(getchar, NONSENSE)
            if retval != NONSENSE:
                return dt.try_raising(retval)

    @staticmethod
    def clear():
        """Clears the screen by click.clear() which is OS independent."""
        click.clear()

    @staticmethod
    def paged_display(source, sep='\n'):
        """Display paginated text so that the user can scroll through it;
        works with any iterable"""
        if not source:
            return
        text = sep.join(str(x) for x in source)
        click.echo_via_pager(text)

    def display(self, *args, sep=' ', end='\n', file=None, min_verbosity=0):
        """Displays info for the user:
        args : iterable of arguments to display,
        sep : separation string (default: space),
        end : termination string (default: newline),
        file : redirect output to specified file,
        min_verbosity=0 : display only if this verbosity is met or exceeded
                          (useful for debug-only output)"""
        if self.verbosity >= min_verbosity:
            line = sep.join(str(arg) for arg in args)
            click.echo(message='{}{}'.format(line, end), nl=False, file=file)

    @staticmethod
    def display_header(text, symbol='-', trailing_newline=1):
        """Displays a header banner.

        text : header text,
        symbol : line symbol,
        trailing_newline : a number of \n (newline) characters
                           after the header.
        """
        header = '{line}\n{text}\n{line}' + '\n' * trailing_newline
        click.echo(header.format(text=text, line=symbol * len(text)))

    def display_parameters(self, *data):
        """Takes any number of OrderedDict instances (for keeping the
        item order intact) and iterates over them, displaying parameters.

        data : [OrderedDict('': header, par1: val1, par2: val2)...]

        If key evaluates False, display a header."""
        for parameter_dict in data:
            for name, value in parameter_dict.items():
                if name:
                    # parameter : value
                    entry = '{name} : {value}'.format(name=name, value=value)
                    click.echo(entry)
                else:
                    # parameter evaluating False: value is a header
                    self.display_header('{}'.format(value), trailing_newline=0)

    def pause(self, msg1='', msg2='Press any key to continue...',
              min_verbosity=0, allow_abort=False):
        """Waits until user presses a key"""
        if self.verbosity >= min_verbosity:
            abort_text = ''
            abort_key_chars = []
            if allow_abort:
                keys = ', '.join(key.name for key in DEFAULT_ABORT_KEYS)
                abort_text = click.style(' [{}: abort]'.format(keys),
                                         fg='cyan')
                abort_key_chars = [key.getchar for key in DEFAULT_ABORT_KEYS]
            _msg2 = click.style(msg2, fg='cyan', blink=True)
            click.echo('{}\n\t{}{}'.format(msg1, _msg2, abort_text))
            char = click.getchar()
            if char in abort_key_chars:
                raise Abort

    @staticmethod
    def edit(text=''):
        """Use click to call a text editor for editing a text"""
        try:
            edited = click.edit(text, editor='nano -t', require_save=False)
        except click.ClickException:
            edited = click.edit(text, require_save=False)
        except click.Abort:
            return text
        return edited

    @staticmethod
    def enter(prompt='Enter the value',
              default=None, datatype=None,
              minimum=None, maximum=None,
              condition=lambda x: True,
              allow_abort=True, type_prompt=None):
        """Enter data based on function arguments.

        prompt :    set the custom prompt to show to the user,

        default :   If a default value is given, it will be already filled
                    in the input, and returned if nothing is entered.

                    If this is an exception, empty input will raise it.

                    If None, no empty value is allowed and user will be asked
                    until a proper value is given.

        datatype :  allows to override the default datatype.

                    If it is specified, the function will try to convert
                    the user input string to this datatype, and force
                    re-entering the data if validation fails.

                    Defaults to the type of default_value.
                    If default_value is None or an exception,
                    the function returns a string.

                    If default_value is False and no datatype is specified,
                    the function returns False.

        minimum, maximum : validation limit values;
                           applies to numeric value (float, int),
                           or string/container's length.

        condition: any additional condition check the value needs to
                   pass in validation.

        allow_abort : if ctrl-C is pressed, the function will
                      raise Abort to be handled elsewhere.

        type_prompt : info for user about what to enter; if None, use default
        """
        def build_prompt():
            """Make a prompt about required datatype and limits."""
            retval_type_handler = dt.get_handler(retval_datatype)

            # what type should the input be?
            type_name = retval_type_handler.type_name
            if type_prompt:
                type_text = type_prompt
            elif type_name:
                type_text = 'Type: {} '.format(type_name)
            else:
                type_text = ''

            # what limits are imposed?
            validated_parameter = retval_type_handler.validated_parameter
            vp_name = dt.LIMITED_PARAMETERS.get(validated_parameter)
            min_string = 'min: {}'.format(minimum)
            max_string = 'max: {}'.format(maximum)
            limits_string = [min_string if minimum is not None else '',
                             max_string if maximum is not None else '']
            limits_str = ', '.join(x for x in limits_string if x)
            limits_text = ('({} {}) '.format(vp_name, limits_str) if limits_str
                           else '')
            requirements = '{}{}'.format(type_text, limits_text)
            # glue it all together
            if requirements:
                return '{}\n{}'.format(prompt, requirements)
            else:
                return prompt

        def get_user_input():
            """Enter the value and return it"""
            # get a prefill value from type definition
            # prefill takes a default and if it evaluates to False
            def prefill_callback():
                """A insert_text function wrapper"""
                pf_value = dt.get_string(default, default_value_type)
                return readline.insert_text(pf_value)

            # get value from user
            readline.set_startup_hook(prefill_callback)
            que = '[Ctrl-C = abort] ?: ' if allow_abort else '> ?: '
            value = input(click.style(que, fg='cyan'))
            return value

        # desired type:
        # specified datatype -> type of default value -> string
        default_value_type = dt.get_type(default)
        retval_datatype = datatype or default_value_type or str
        # configure the conversion/validation function
        conv_validate = partial(dt.convert_and_validate,
                                default=default,
                                datatype=retval_datatype,
                                minimum=minimum, maximum=maximum,
                                condition=condition)

        # tell user about the constraints
        click.echo(build_prompt())

        # loop until the function returns correct value
        while True:
            # get the value from wrapped function
            # raise exceptions (if default value was an exception)
            try:
                value = conv_validate(get_user_input())
                if value is None:
                    continue
                else:
                    return value

            except (TypeError, ValueError) as error:
                # show the message and loop again
                click.secho('Error: {}'.format(error), fg='red')
            except (KeyboardInterrupt, EOFError, click.Abort):
                if allow_abort:
                    raise Abort

            finally:
                click.echo('\n')
                readline.set_startup_hook()

    @staticmethod
    def open_file(default_filename='',
                  mode='r', message='File name?',
                  allow_abort=True):
        """Allows to enter the input filename and checks if it is readable.
        Repeats until proper filename or nothing is given.
        Returns a file object.
        Raises Abort if filename not specified."""
        def readline_prefill():
            """Pre-fill the input prompt for readline."""
            return readline.insert_text(default_filename)

        # Set readline parameters
        readline.set_completer_delims(' \t\n;')
        readline.parse_and_bind('tab: complete')
        readline.set_completer(lambda text, state:
                               (glob.glob(text+'*')+[None])[state])
        if allow_abort:
            abort_text = click.style('[leave blank to abort]', fg='cyan')
            prompt = '{} {} ?: '.format(message, abort_text)
        else:
            prompt = '{}  ?:'.format(message)
        # Check file parameters and catch early permission errors
        checks = {'r': {'readable': True, 'exists': True},
                  'w': {'writable': True, 'exists': True},
                  'w+': {'writable': True, 'exists': False}}
        while True:
            try:
                readline.set_startup_hook(readline_prefill)
                click.echo()
                filename = input(prompt)
                click.echo()
                if not filename and allow_abort:
                    raise Abort
                # okay, got a path, check if file can be accessed
                path = click.Path(resolve_path=True, **checks[mode])(filename)
                return click.File(mode)(path)
            except (click.BadParameter, click.FileError) as exc:
                # permission denied and wrong file exception handler
                click.secho(exc.message, fg='red')
            except (KeyboardInterrupt, EOFError):
                # ctrl-C exception handler
                if allow_abort:
                    raise Abort
            finally:
                # readline prefill cleanup
                readline.set_startup_hook()

    def import_file(self, default_filename='', allow_abort=True):
        """Allows user to enter the input filename. Returns a file object.
        Returns default filename or raises Abort if filename not specified."""
        return self.open_file(default_filename,
                              mode='r', message='Enter the input filename',
                              allow_abort=allow_abort)

    def export_file(self, default_filename='', allow_abort=True):
        """Allows user to enter the output filename. Returns a file object.
        Returns default filename or raises Abort if filename not specified."""
        return self.open_file(default_filename,
                              mode='w+', message='Enter the output filename',
                              allow_abort=allow_abort)

    @staticmethod
    def confirm(question='Your choice?', default=True,
                abort_answer=None, allow_abort=True):
        """Asks a simple question with yes or no answers.
        Returns True for yes and False for no.

        default : default answer if user presses return,
        abort_answer : if True or False, yes / no answer raises Abort;
                       if None, the outcome is returned for both answers.

        allow_abort : allows aborting by ctrl-C or Esc
        """
        # key definitions and their meanings
        keys = OrderedDict()
        keys[get_key('y')] = True
        keys[get_key('n')] = False
        keys[get_key('esc')] = Abort if allow_abort else False

        names = {True: 'yes', False: 'no'}

        default_text, abort_text = '', ''
        prompt = click.style('Choice?', fg='yellow', bold=True)
        yn_text = click.style('Y = yes, N = no', fg='cyan', bold=True)

        # default and abort answer
        if default is not None:
            keys[get_key('enter')] = default
            default_text = click.style(' Enter = {}'.format(names[default]),
                                       fg='green', bold=True)

        if allow_abort:
            keys[get_key('esc')] = Abort
            abort_text = click.style(' Esc = abort', fg='red', bold=True)

        # all keys are defined
        # build answer dict from key getchars
        answers = {key.getchar: answer for key, answer in keys.items()}
        # display user prompts
        click.echo(question)

        while True:
            # get the user input
            click.echo('{} [{}{}{}]'
                       .format(prompt, yn_text, default_text, abort_text))
            getchar = click.getchar()
            answer = answers.get(getchar)
            if answer == abort_answer:
                raise Abort

            # loop further if answer lookup failed
            if answer is None:
                continue

            # return answer, or raise it (if it was Abort)
            return dt.try_raising(answer)

    def paused(self, routine):
        """Execute a routine and pause after it's done"""
        def wrapper(*args, **kwargs):
            """wrapper function"""
            retval = routine(*args, **kwargs)
            self.pause()
            return retval

        return wrapper



# -*- coding: utf-8 -*-
"""Basic controller routines - elementary setters/getters.
Depend on models. Used by higher-order controllers."""
from contextlib import suppress
from functools import partial, wraps
from itertools import zip_longest
from .rpi2caster import CFG, UI, Abort, Finish, option
from . import basic_models as bm, definitions as d


# Letter frequency controller routines

def define_scale(freqs):
    """Define scale of production and upper.lowercase ratio"""
    prompt = ('How much lowercase "a" characters do you want to cast?\n'
              'The quantities of other characters will be calculated\n'
              'based on the letter frequency of the language.\n'
              'Minimum of 10 characters each will be cast.')
    freqs.scale = UI.enter(prompt, default=100, datatype=int)
    prompt = 'Uppercase to lowercase ratio in %?'
    freqs.case_ratio = UI.enter(prompt, default=20.0, datatype=float) / 100.0


def get_letter_frequencies():
    """Display available languages and let user choose one.
    Returns a string with language code (e.g. en, nl, de)."""
    langs = ('{} - {}'.format(c, lang) for c, lang in sorted(d.LANGS.items()))

    # Group by three
    grouper = zip_longest(langs, langs, langs, fillvalue='')
    descriptions = '\n'.join('\t'.join(z) for z in grouper)

    UI.display('Choose language:\n\n{}\n'.format(descriptions))
    prompt = 'Language code (e.g. "en")'
    while True:
        lang = UI.enter(prompt, default=Abort, datatype=str, maximum=2)
        try:
            return bm.CharFreqs(lang)
        except KeyError:
            UI.display('Language {} not found. Choose again.'.format(lang))


# Matrix control routines

def edit_matrix(matrix, single=True):
    """Edits the matrix data.

    matrix : a Matrix class object to edit,

    edit_char : whether or not to edit the character,
    edit_position : whether or not to edit the matrix coordinates,
    edit_styles : whether or not to change the styles for the matrix,
    edit_units : whether or not to change the matrix unit value
    single : if True, editing only one matrix, else: more
             (adds an option to escape editing the whole series)
    """
    def _edit_char():
        """Edit the matrix character"""
        prompt = 'Character? (blank = keep current)'
        matrix.char = UI.enter(prompt, default=matrix.char or '')
        # if this is a space, then done
        if matrix.isspace():
            return
        # change styles
        matrix.styles = choose_styles(matrix.styles,
                                      mask=matrix.diecase.styles)

    def _edit_dimensions():
        """Edit the matrix size"""
        current_size = matrix.size
        options = [option(key='n', value=(1, 1), seq=1, text='1x1 - normal'),
                   option(key='w', value=(2, 1), seq=2, text='2x1 - wide'),
                   option(key='h', value=(1, 2), seq=3, text='1x2 - high'),
                   option(key='l', value=(2, 2), seq=4, text='2x2 - large'),
                   option(key='Enter', value=current_size, seq=5,
                          text='keep current ({})'.format(current_size))]
        new_size = UI.simple_menu('New matrix size?', options,
                                  allow_abort=False)
        matrix.size = d.MatrixSize(new_size)

    def _edit_position():
        """Edit the matrix coordinates"""
        def move_outside(new, current):
            """Move the old mat to outside characters, and replace it with
            a new mat"""
            current.code = None
            diecase.outside_mats.append(current)
            new.code = code

        def swap(new, current):
            """Swap two matrices"""
            new.code, current.code = current.code, new.code

        code = UI.enter('New position for this matrix?',
                        default=matrix.code or '').upper()
        try:
            # look for a mat in the target position
            diecase = matrix.diecase
            if not code:
                diecase.outside_mats.append(matrix)
            current_mat = diecase.select_one(code=code)
            if current_mat:
                UI.display('There is already a mat in the target position.')
                question = 'What to do with {}?'.format(current_mat)
                options = [option('s', swap, text='Swap the matrices', seq=1),
                           option('o', move_outside, seq=1,
                                  text='Remove the current matrix'),
                           option('Esc', Abort, text='Cancel')]
                UI.simple_menu(question, options)(matrix, current_mat)
            else:
                matrix.code = code
        except bm.MatrixNotFound:
            # no mat in the target position = move the mat freely
            matrix.code = code
        except Abort:
            pass

    def _edit_units():
        """Change the matrix unit width value"""
        # skip this for spaces
        if not _edit_units:
            return

        # get unit width values
        row_units = matrix.get_units_from_row()
        ua_units = matrix.get_units_from_arrangement()
        curr_units = matrix.units

        # build a prompt with units info
        curr_chunk = '' if not curr_units else 'current: {}'.format(curr_units)
        row_chunk = '' if not row_units else 'row units: {}'.format(row_units)
        ua_chunk = '' if not ua_units else 'UA units: {}'.format(ua_units)
        chunks = [x for x in (curr_chunk, row_chunk, ua_chunk) if x]
        if chunks:
            prompt = 'Enter unit width ({})'.format(', '.join(chunks))
        else:
            prompt = 'Enter unit width'
        new_units = UI.enter(prompt, default=curr_units, datatype=int,
                             minimum=3, maximum=30)
        matrix.units = new_units or ua_units or row_units

    def options():
        """Generate menu options"""
        ret = [option(key='c', value=_edit_char, seq=1,
                      text='Set the character and assign styles'),
               option(key='l', value=partial(matrix.set_lowspace, True),
                      text='Set the matrix as low space', seq=10,),
               option(key='h', value=partial(matrix.set_highspace, True),
                      text='Set the matrix as high space', seq=11),
               option(key='p', value=_edit_position, seq=20,
                      text='Move the matrix to a different position'),
               option(key='w', value=_edit_units, seq=40,
                      cond=not matrix.isspace(), text='Change the unit width'),
               option(key='d', value=_edit_dimensions, seq=50,
                      text=('Change the matrix size (current: {}x{} cells)'
                            .format(*matrix.size))),
               option(key='Enter', value=Abort, seq=90,
                      text='Finish editing the matrix', cond=single),
               option(key='Enter', value=Abort, seq=90,
                      text='Edit the next matrix in series', cond=not single),
               option(key='Esc', value=Finish, seq=99, cond=not single,
                      text='Return to the layout editor menu')]
        return ret

    with suppress(Abort):
        # keep displaying this menu until aborted
        while True:
            # display the menu for user to choose
            header = 'Edit the matrix for {}:'.format(matrix)
            choice = UI.simple_menu(header, options, allow_abort=False)
            # execute the subroutine
            choice()
    return matrix


# Measure controller routines

def set_measure(input_value=CFG['Typesetting'].get('default_measure'),
                unit=CFG['Typesetting'].get('measurement_unit'),
                what='measure', set_width=12.0):
    """Enter the line length, choose measurement units
    (for e.g. British or European measurement system).
    Return length in DTP points."""
    info = ('\n\nAvailable units:\n'
            '    dd - European Didot point,\n'
            '    cc - cicero (=12dd, .1776"),\n'
            '    ff - Fournier point,\n'
            '    cf - Fournier cicero (=12ff, .1628"),\n'
            '    pp - TeX / US printer\'s pica point,\n'
            '    Pp - TeX / US printer\'s pica (=12pp, .1660"),\n'
            '    pt - DTP / PostScript point = 1/72",\n'
            '    pc - DTP / PostScript pica (=12pt, .1667"),\n'
            '    em - 18 units, en - 9 units, u - 1 unit {} set,\n'
            '    ", in - inch;   mm - millimeter;   cm - centimeter\n\n')

    prompt = 'Enter the {} value and unit (or "?" for help)'.format(what)
    while True:
        # If 0, use default
        raw_string = UI.enter(prompt, default='{}'.format(input_value))
        if '?' in raw_string:
            # Display help message and start again
            UI.display(info.format(set_width))
            continue
        try:
            return bm.Measure(raw_string, unit, set_width)
        except (TypeError, ValueError):
            UI.display('Incorrect value - enter again...')


def temp_measure(routine):
    """Decorator for casting and typesetting routines.
    Allow user to change measure i.e. line length"""
    @wraps(routine)
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        description = 'galley width (line length)'
        # Change measure before, restore after
        old_measure = self.measure
        self.measure = set_measure(old_measure, what=description,
                                   set_width=self.wedge.set_width)
        retval = routine(self, *args, **kwargs)
        self.measure = old_measure
        return retval
    return wrapper


# Style controller routines

def choose_styles(styles='', default=None, multiple=True, mask=None):
    """Manual style choice"""
    def select(sty):
        """Selects a style - either one or multiple"""
        if not multiple:
            current.clear()
        current.add(sty)

    def deselect(sty):
        """Remove the style from currently selected"""
        current.discard(sty)

    def header():
        """Build a header"""
        text = ', '.join(s.name for s in bm.Styles.definitions if s in current)
        names = text or '{} (default)'.format(default_style.name)
        what = 'one or more text styles' if multiple else 'a text style'
        return 'Select {}.\nCurrently selected: {}'.format(what, names)

    def options():
        """Make menu entries for style selection menu"""
        items = []
        number = 1
        for sty in bm.Styles.definitions:
            text_on = 'Select {}'.format(sty.name)
            st_on = option(key=sty.short, text=text_on, seq=number,
                           value=partial(select, sty),
                           cond=sty in _mask and sty not in current)
            text_off = 'Deselect {}'.format(sty.name)
            st_off = option(key=sty.short, text=text_off, seq=number,
                            value=partial(deselect, sty),
                            cond=multiple and sty in current)
            items.extend([st_on, st_off])
            number += 1
        items.append(option(key='enter', text='Finish', value=Finish))
        return items

    # make a mask limiting the choice of styles
    _mask = bm.Styles(mask or '*')
    # what styles are currently chosen?
    default_style = bm.STYLES.get(default, d.STYLES.roman)
    current = {s for s in bm.Styles(styles, default=default_style)}
    try:
        while True:
            UI.simple_menu(header, options, allow_abort=True)()
    except Finish:
        return bm.Styles(current, default_style, _mask)



# -*- coding: utf-8 -*-
"""
    rpi2caster is a CAT (computer-aided typesetting) software
    for the Monotype composition caster (a hot-metal typesetting machine).

    This project uses a control interface for 31/32 solenoid valves
    and a machine cycle sensor. It can control a casting machine
    or a pneumatic paper tape perforator from the Monotype keyboard.

    The rpi2caster package consists of three main utilities:

        * casting (composition, material etc.),

        * caster/interface testing and diagnostics,

        * typesetting (not ready yet),

        * diecase and layout management.

    Machine control utility also serves as a diagnostic program
    for calibrating and testing the machine and control interface.

"""
from collections import OrderedDict
from functools import partial
import json
import os
from pathlib import Path

import click
import librpi2caster

from . import data, global_state
from .ui import Abort, Finish, option

# get singleton instances for user interface, database and configuration
USER_DATA_DIR = global_state.USER_DATA_DIR
UI = global_state.UI
DB = global_state.DB
CFG = global_state.CFG


class CommandGroup(click.Group):
    """Click group which allows using abbreviated commands,
    and arranges them in the order they were defined."""
    def __init__(self, name=None, commands=None, **attrs):
        if commands is None:
            commands = OrderedDict()
        elif not isinstance(commands, OrderedDict):
            commands = OrderedDict(commands)
        click.Group.__init__(self, name=name, commands=commands, **attrs)

    def list_commands(self, ctx):
        """List command names as they are in commands dict."""
        return self.commands.keys()

    def get_command(self, ctx, cmd_name):
        """Try to get a command with given partial name;
        in case of multiple match, abort."""
        retval = click.Group.get_command(self, ctx, cmd_name)
        if retval is not None:
            return retval
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


def find_casters(operation_mode=None):
    """Finds casters and returns dictionary of number-caster values"""
    from .monotype import MonotypeCaster, SimulationCaster

    def make_caster(url):
        """caster factory method: make a real or simulation caster;
        if something bad happens, just return None"""
        try:
            caster = MonotypeCaster(url, operation_mode)
            return (caster, str(caster))
        except librpi2caster.InterfaceException as exc:
            return (None, str(exc))

    # get the interface URLs
    # the first interface is a simulation interface numbered 0
    config_urls = CFG['System']['interfaces']
    caster_urls = [*(x.strip() for x in config_urls.split(','))]
    # make a dictionary of casters starting with 0 for a simulation caster
    casters = {0: ('', SimulationCaster(), 'Simulation mode - no hardware')}
    for number, url in enumerate(caster_urls, start=1):
        caster, name = make_caster(url)
        casters[number] = (url, caster, name)
    return casters


def add_extra(raw_path, target):
    """Read the JSON file from path; decode the data; update the target."""
    path = Path(raw_path)
    source = path if path.is_absolute() else Path(USER_DATA_DIR).joinpath(path)
    try:
        file = source.resolve()
        with file.open('r') as jsonfile:
            extra_data = json.load(jsonfile)
            target.update(extra_data)
    except FileNotFoundError:
        UI.display('File {} cannot be found, skipping.'.format(source),
                   min_verbosity=1)
    except (IOError, json.JSONDecodeError):
        UI.display('Cannot read file {}, skipping.'.format(source),
                   min_verbosity=1)


@click.group(invoke_without_command=True, cls=CommandGroup, help=__doc__,
             context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(None, '--version', '-V')
@click.option('verbosity', '-v', count=True, default=0,
              help='verbose mode (count, default=0)')
@click.option('--conffile', '-c', help='config file to use', metavar='[PATH]',
              default=os.path.join(USER_DATA_DIR, 'rpi2caster.conf'))
@click.option('--database', '-d', metavar='[URL]', help='database URL to use')
@click.option('--web', '-W', 'ui_impl', flag_value='web_ui',
              help='use web user interface (not implemented)')
@click.option('--text', '-T', 'ui_impl', flag_value='text_ui',
              default=True, help='use text user interface')
@click.pass_context
def cli(ctx, conffile, database, ui_impl, verbosity):
    """decide whether to go to a subcommand or enter main menu"""
    def menu_options():
        """Dynamically generate options"""
        ret = [option(key='t', value=partial(ctx.invoke, translate), seq=10,
                      text='Typesetting...',
                      desc='Compose text for casting'),

               option(key='c', value=partial(ctx.invoke, cast), seq=20,
                      text='Casting or punching...',
                      desc=('Cast composition, sorts, typecases or spaces;'
                            ' test the machine')),

               option(key='d', value=partial(ctx.invoke, edit_diecase), seq=30,
                      text='Diecase manipulation...',
                      desc='Manage the matrix case collection'),

               option(key='u', value=partial(ctx.invoke, update),
                      text='Update the program', seq=90)]
        return ret

    CFG.read(conffile)
    # get the URL from the argv or updated config
    database_url = database or CFG['System']['database']
    DB.load(database_url)
    UI.load(ui_impl, verbosity)

    # read the additional typeface/wedge/unit_arrangement data
    add_extra(CFG['System']['extra_typefaces'], data.TYPEFACES)
    add_extra(CFG['System']['extra_unit_arrangements'], data.UNIT_ARRANGEMENTS)
    add_extra(CFG['System']['extra_wedges'], data.WEDGE_DEFINITIONS)
    add_extra(CFG['System']['extra_languages'], data.LETTER_FREQUENCIES)

    # main menu
    header = ('rpi2caster - computer aided typesetting software '
              'for Monotype composition casters.'
              '\n\nMain menu:\n')
    exceptions = (Abort, Finish, click.Abort)
    if not ctx.invoked_subcommand:
        UI.dynamic_menu(menu_options, header, allow_abort=True,
                        catch_exceptions=exceptions)


@cli.group(invoke_without_command=True, cls=CommandGroup,
           options_metavar='[-hlmsw]', subcommand_metavar='[what] [-h]')
@click.option('--interface', '-i', default=None, type=int, metavar='[number]',
              help='choose interface:\n0=simulation, 1,2...=hardware')
@click.option('--punching', '-p', 'operation_mode', flag_value='punching',
              help='punch ribbon with a perforator (if supported)')
@click.option('--casting', '-c', 'operation_mode', flag_value='casting',
              help='cast type on a composition caster (if supported)')
@click.option('--diecase', '-m', metavar='[diecase ID]',
              help='diecase ID from the database to use')
@click.option('--wedge', '-w', metavar='e.g. S5-12E',
              help='series, set width, E for European wedges')
@click.option('--measure', '-l', metavar='[value+unit]',
              help='line length to use')
@click.pass_context
def cast(ctx, interface, operation_mode, diecase, wedge, measure):
    """Cast type with a Monotype caster.

    Casts composition, material for handsetting, QR codes.
    Can also cast a diecase proof.

    Can also be run in simulation mode without the actual caster."""
    from .core import Casting
    # allow override if we call this from menu
    casting = Casting(interface, operation_mode)
    casting.measure = measure
    casting.diecase_id = diecase
    casting.wedge_name = wedge
    # replace the context object for the subcommands to see
    ctx.obj = casting
    if not ctx.invoked_subcommand:
        casting.main_menu()


@cast.command('ribbon', options_metavar='[-h]')
@click.argument('ribbon', metavar='[filename|ribbon_id]')
@click.pass_obj
def cast_ribbon(casting, ribbon):
    """Cast composition from file or database."""
    try:
        casting.ribbon_by_name(ribbon)
        casting.cast_composition()
    except FileNotFoundError:
        UI.display('File {} not found.'.format(ribbon))


@cast.command('material', options_metavar='[-h]')
@click.pass_obj
def cast_handsetting_material(casting):
    """Cast founts, sorts and spaces/quads."""
    casting.cast_material()


@cast.command('qrcode', options_metavar='[-h]')
@click.pass_obj
def cast_qr_code(casting):
    """Generate and cast QR codes."""
    casting.cast_qr_code()


@cast.command('proof')
@click.pass_obj
def cast_diecase_proof(casting):
    """Cast a matrix case proof."""
    casting.diecase_proof()


@cast.command('test', options_metavar='[-hps]')
@click.pass_obj
def test_machine(casting):
    """Monotype caster testing and diagnostics."""
    casting.machine.diagnostics_menu()


@cli.command(options_metavar='[-ahlmMw] [--src textfile] [--out ribbonfile]')
@click.option('--src', type=click.File('r'))
@click.option('--out', type=click.File('w+', atomic=True))
@click.option('--diecase', '-m', metavar='[diecase ID]',
              help='diecase ID from the database to use')
@click.option('--wedge', '-w', metavar='e.g. S5-12E',
              help='series, set width, E for European wedges')
@click.option('--measure', '-l', metavar='[value+unit]',
              help='line length to use')
@click.option('--align', '-a', metavar='[ALIGNMENT]', default='left',
              help='default text alignment')
@click.option('--manual', '-M', is_flag=True, flag_value=True,
              help='leave end-of-line decisions to the operator')
def translate(src, out, align, manual, **kwargs):
    """Typesetting program.

    Set and justify a text, using control codes for styles, alignment etc.

    The output is a sequence of codes which can control the
    Monotype composition caster.

    These codes are specific to a diecase and wedge used."""
    from .core import Typesetting
    typesetting = Typesetting()
    typesetting.measure = kwargs.get('measure')
    typesetting.diecase_id = kwargs.get('diecase')
    typesetting.wedge_name = kwargs.get('wedge')
    typesetting.manual_mode = manual
    typesetting.default_alignment = align
    typesetting.text_file = src
    typesetting.ribbon.file = out
    # Only one method here
    typesetting.main_menu()


@cli.command('edit', options_metavar='[-h]')
@click.argument('diecase', required=False, default=None,
                metavar='[diecase_id]')
def edit_diecase(diecase):
    """Load and edit a matrix case."""
    from . import main_controllers as mc
    editor = mc.DiecaseMixin()
    editor.diecase_id = diecase
    editor.diecase_manipulation()


@cli.group('list', invoke_without_command=True, cls=CommandGroup,
           options_metavar='[-h]', subcommand_metavar='[d|r|t|u|w] [-h]')
@click.pass_context
def _list(ctx):
    """List items found in database or definitions"""
    if not ctx.invoked_subcommand:
        ctx.invoke(list_diecases)


@_list.command('diecases', options_metavar='[-h]')
def list_diecases():
    """List all available diecases and exit."""
    from . import views, main_controllers as mc
    views.list_diecases(mc.get_all_diecases())


@_list.command('wedges', options_metavar='[-h]')
def list_wedges():
    """List all known wedge definitions, and exit."""
    from . import views
    views.list_wedges()


@_list.command('typefaces', options_metavar='[-h]')
def list_typefaces():
    """List all known typefaces and exit."""
    from . import views
    views.list_typefaces()


@_list.command('uas', options_metavar='[-h]')
def list_uas():
    """List all known unit arrangements and exit."""
    from . import views
    views.list_unit_arrangements()


@_list.command('machines', options_metavar='[-h]')
@click.option('--punching', '-p', 'mode', flag_value='punching',
              help='punch ribbon with a perforator (if supported)')
@click.option('--casting', '-c', 'mode', flag_value='casting',
              help='cast type on a composition caster (if supported)')
def list_machines(mode):
    """List all configured casters and show the available ones."""
    machines = find_casters(mode)
    UI.display('\nList of configured interfaces for {}:\n'
               .format(mode or 'casting or punching'))
    for url, caster, name in machines.values():
        if not url:
            # don't list the simulation interface - it's always available
            continue
        url_string = ('{}'.format(url) if caster
                      else '{}: unavailable'.format(url))
        if len(str(name)) > 30:
            template = '{}\n{}\n'
        else:
            template = '{} :\t{}\n'
        UI.display(template.format(url_string, name))


@cli.group(cls=CommandGroup, options_metavar='[-h]',
           subcommand_metavar='[d|l|r|t|u|w] [-h]')
def show():
    """Show an item (diecase, layout, UA etc.)"""


@show.command('layout', options_metavar='[-h]')
@click.argument('diecase', required=False, default=None,
                metavar='[diecase_id]')
def show_layout(diecase):
    """Display a diecase layout for a specified diecase ID."""
    from . import views, main_controllers as mc
    case = mc.get_diecase(diecase)
    views.display_layout(case)


@show.command('diecase', options_metavar='[-h]')
@click.argument('diecase', required=False, default=None,
                metavar='[diecase_id]')
def show_diecase_data(diecase):
    """Display diecase data."""
    from . import views, main_controllers as mc
    case = mc.get_diecase(diecase)
    views.display_diecase_data(case)


@show.command('typeface', options_metavar='[-h]')
@click.argument('typeface', metavar='[number or name]')
def show_typeface(typeface):
    """Find a typeface and show its data."""
    from . import views
    views.display_typeface(typeface)


@show.command('ua', options_metavar='[-h]')
@click.argument('number', metavar='[UA number]')
def show_unit_arrangement(number):
    """Shows the UA data for the given UA ID."""
    from . import views
    views.display_ua(number)


@show.command('wedge', options_metavar='[-h]')
@click.argument('designation', metavar='[wedge name]')
def show_wedge(designation):
    """Shows the wedge parameters."""
    from . import views
    views.display_wedge(designation)


@cli.group(invoke_without_command=True, cls=CommandGroup,
           options_metavar='[-h]', subcommand_metavar='[d|r] [-h]')
@click.pass_context
def options(ctx):
    """Display the local and global configuration for rpi2caster."""
    if not ctx.invoked_subcommand:
        ctx.invoke(options_dump)


@options.command('dump', options_metavar='[-h]')
@click.argument('output', required=False, type=click.File('w+'),
                default=click.open_file('-', 'w'), metavar='[path]')
def options_dump(output):
    """Write all settings to a file or stdout.

    Dumps all current configuration settings, both user-specific and global.

    Path can be any file on the local filesystem."""
    CFG.write(output)


@options.command('read', options_metavar='[-h]')
@click.argument('cfg_option')
def options_read(cfg_option):
    """Read a specified option from configuration.

    If option is not found, display error message."""
    for section_name in CFG.sections():
        section = CFG[section_name]
        for option_name, option_value in section.items():
            if option_name.lower() == cfg_option.lower():
                UI.display_header(section)
                UI.display(option_value)


@cli.command(options_metavar='[-ht]')
@click.option('--testing', '-t', is_flag=True, flag_value=True,
              help='use a unstable/development version instead of stable')
def update(testing):
    """Update the software."""
    # Upgrade routine
    dev_prompt = 'Testing version (newest features, but unstable)? '
    if UI.confirm('Update the software?', default=False):
        use_dev_version = testing or UI.confirm(dev_prompt, default=False)
        pre = '--pre' if use_dev_version else ''
        os.system('pip3 install {} --upgrade rpi2caster'.format(pre))


@cli.command()
def meow():
    "Easter egg."
    try:
        UI.display('\nOh, this was meowsome.\n')
        UI.display(data.EASTER_EGG)
    except (OSError, ImportError, FileNotFoundError):
        print('There are no Easter Eggs in this program.')



# -*- coding: utf-8 -*-
"""Functions and classes for parsing strings/iterables for usable data."""
from collections import namedtuple
from contextlib import suppress
from functools import lru_cache
import re

# parsing delimiters
COMMENT_SYMBOLS = ['**', '*', '//', '##', '#']
ASSIGNMENT_SYMBOLS = ['=', ':', ' ']

# definitions
Token = namedtuple('Token', 'name value routine')
ParsedRecord = namedtuple('ParsedRecord',
                          ('raw signals comment column row has_signals is_char'
                           ' uses_row_16 has_0005 has_0075 has_s_needle '
                           'is_pump_start is_pump_stop is_newline'))


@lru_cache(maxsize=350)
def parse_record(input_data):
    """Parses the record and gets its row, column and justification codes.
    First split the input data into two parts:
    -the Monotype signals (unprocessed),
    -any comments delimited by symbols from COMMENT_SYMBOLS list.

    Looks for any comment symbols defined here - **, *, ##, #, // etc.
    splits the line at it and saves the comment to return it later on.
    If it's an inline comment (placed after Monotype code combination),
    this combination will be returned for casting."""
    def split_on_delimiter(sequence):
        """Iterate over known comment delimiter symbols to find
        whether a record has a comment; split on that delimiter
        and normalize the signals"""
        source = ''.join(str(x) for x in sequence)
        for symbol in COMMENT_SYMBOLS:
            if symbol in source:
                # Split on the first encountered symbol
                raw_signals, comment = source.split(symbol, 1)
                break
        else:
            # no comment symbol encountered, so we only have signals
            raw_signals, comment = source, ''

        return raw_signals.upper().strip(), comment.strip()

    def find(value):
        """Detect and dispatch known signals in source string"""
        nonlocal signals
        string = str(value)
        if string in signals:
            signals = signals.replace(string, '')
            return True
        else:
            return False

    def analyze():
        """Check if signals perform certain functions"""
        has_signals = any((columns, rows, justification))
        column = (None if not has_signals
                  else 'NI' if set('NI').issubset(columns)
                  else 'NL' if set('NL').issubset(columns)
                  else columns[0] if columns else 'O')
        row = None if not has_signals else rows[0] if rows else 15
        has_0005 = '0005' in justification or set('NJ').issubset(columns)
        has_0075 = '0075' in justification or set('NK').issubset(columns)
        uses_row_16 = row == 16
        is_pump_start = has_0075
        is_pump_stop = has_0005 and not has_0075
        is_newline = has_0005 and has_0075
        is_char = not justification
        has_s_needle = 'S' in signals

        return dict(has_signals=has_signals, column=column, row=row,
                    has_0005=has_0005, has_0075=has_0075, is_char=is_char,
                    is_newline=is_newline, uses_row_16=uses_row_16,
                    is_pump_start=is_pump_start, is_pump_stop=is_pump_stop,
                    has_s_needle=has_s_needle)

    # we know signals and comment right away
    raw_signals, comment = split_on_delimiter(input_data)
    signals = raw_signals

    # read the signals to know what's inside
    justification = tuple(x for x in ('0075', '0005') if find(x))
    parsed_rows = [x for x in range(16, 0, -1) if find(x)]
    rows = tuple(x for x in reversed(parsed_rows))
    columns = tuple(x for x in [*'ABCDEFGHIJKLMNO'] if find(x))

    return ParsedRecord(signals=raw_signals, comment=comment,
                        raw=input_data, **analyze())


def parse_ribbon(ribbon):
    """Get the metadata and contents out of a sequence of lines"""
    def get_value(line, symbol):
        """Helper function - strips whitespace and symbols"""
        # Split the line on an assignment symbol, get the second part,
        # strip any whitespace or multipled symbols
        return line.split(symbol, 1)[-1].strip(symbol).strip()

    # What to look for
    keywords = ['diecase', 'description', 'desc', 'diecase_id', 'customer',
                'wedge', 'stopbar']
    targets = ['diecase_id', 'description', 'description', 'diecase_id',
               'customer', 'wedge_name', 'wedge_name']
    parameters = dict(zip(keywords, targets))
    # Metadata (anything found), contents (the rest)
    metadata = {}
    contents = []
    # Look for parameters line per line, get parameter value
    # If parameters exhausted, append the line to contents
    for line in ribbon:
        for keyword, target in parameters.items():
            if line.startswith(keyword):
                for sym in ASSIGNMENT_SYMBOLS:
                    if sym in line:
                        # Data found
                        metadata[target] = get_value(line, sym)
                        break
                break
        else:
            contents.append(line)
    # We need to add contents too
    metadata['contents'] = contents
    return metadata


def tokenize(text, token_specification, ignore_case=False):
    """Get tokens from text"""
    routines = {}
    regexes = []
    for item in token_specification:
        try:
            name, regex, routine = item
            routines[name] = routine
        except ValueError:
            name, regex = item
            routines[name] = name
        regexes.append('(?P<{}>{})'.format(name, regex))

    flags = re.IGNORECASE if ignore_case else 0
    for match in re.finditer('|'.join(regexes), text, flags):
        name = match.lastgroup
        value = match.group(name)
        routine = routines.get(name)
        yield Token(name, value, routine)


def get_code_and_number(pattern):
    """Parse a regex string to get a number"""
    regex = re.compile(r'([0-9]+)')
    print(regex.split(pattern.strip('^')))
    return int(''.join((x for x in pattern if x.isdigit())) or 0)


def make_number_pattern(codes, min_digits=1, max_digits=2):
    """Make a regular expression pattern matching code and number
    from min_digits to max_digits, e.g.
    code: "#", min_digits: 1, max_digits: 3 -> "(#[0-9]{1, 3})"
    """
    return (r'(\^[{c}][0-9]{lb}{minimum},{maximum}{rb})'
            .format(c=re.escape(codes), lb='{', rb='}',
                    minimum=min_digits, maximum=max_digits))


def split_on_matches(text, regex_pattern, ignore_case=True):
    """Split a string on codes. Return a list of chunks:
        text -> [str1, code1, str2, code2...]

        regex_pattern: a custom regular expression pattern to split on,
        ignore_case: no distinction between lowercase and uppercase codes
    """
    flags = re.IGNORECASE if ignore_case else 0
    regex = re.compile(regex_pattern, flags=flags)
    return regex.split(text)


def token_parser(source, *token_sources, skip_unknown=True):
    """Yields tokens (characters, control sequences), one by one,
    as they are found in the source.
    input_stream - iterable;
    skip_unknown - yield only the characters found in token_sources (default),
                   otherwise, unknown characters are also yielded
    token_sources - any number of iterables containing the tokens
                    we are looking for"""
    # Collect all tokens (characters, control sequences) from token_sources
    tokens = [token for sequence in token_sources for token in sequence]
    # Determine the maximum length of a token
    max_len = max(len(t) for t in tokens)
    # We have to skip a number of subsequent input stream characters
    # after a multi-character token is encountered
    skip_steps = 0
    # Characters which will be ignored and not redirected to output
    ignored_tokens = ('\n',)
    # What if char in text not present in diecase? Hmmm...
    for index, _ in enumerate(source):
        if skip_steps:
            # Skip the characters to be skipped
            skip_steps -= 1
            continue
        for i in range(max_len, 0, -1):
            # Start from longest, end with shortest
            with suppress(TypeError, AttributeError, ValueError):
                chunk = source[index:index+i]
                skip_steps = i - 1
                if chunk in ignored_tokens:
                    break
                elif chunk in tokens or i == 1 and not skip_unknown:
                    # Make sure that the function will yield chunks of 1 char
                    yield chunk
                    break



# -*- coding: utf-8 -*-
"""View functions for rpi2caster"""

from collections import OrderedDict
from contextlib import suppress
from itertools import zip_longest
from string import ascii_lowercase, ascii_uppercase, digits

from . import basic_models as bm, definitions as d
from .rpi2caster import UI
from .data import TYPEFACES, UNIT_ARRANGEMENTS
from .data import WEDGE_ALIASES, WEDGE_DEFINITIONS
from .main_models import Typeface, UnitArrangement, UAVariant, Wedge


# Wedge views

def list_wedges():
    """List known wedges"""
    def sorting_key(ua_number):
        """Sorting key: integer part"""
        return int(''.join(char for char in ua_number if char.isdigit()))

    UI.display('Wedge definitions and their unit values:\n\n'
               'The first number is the wedge/stopbar series number.\n'
               'Unit values are listed for each row of the matrix case.\n'
               'Some wedges have an additional 16th unit value. This means '
               'that this wedge was used\nwith a "HMN" or "KMN" early row 16 '
               'addressing system. Those wedges and attachments are rare.\n\n')
    UI.display('The program currently knows {} wedge definitions.'
               .format(len(WEDGE_DEFINITIONS)))
    rows = ''.join(str(x).ljust(5) for x in range(1, 16))
    UI.display_header('{:^10}{}[16]'.format('Wedge', rows))
    # display all wedges
    for wedge_number in sorted(WEDGE_DEFINITIONS, key=sorting_key):
        wedge_units = WEDGE_DEFINITIONS[wedge_number]
        unit_values = ''.join(str(x).ljust(5) for x in wedge_units)
        UI.display('{:^10}{}'.format(wedge_number, unit_values))

    # display old-style wedge designations
    aliases = iter(WEDGE_ALIASES)
    # Group by three
    grouper = zip_longest(aliases, aliases, aliases, fillvalue='')
    old_wedges = '\n'.join('\t'.join(z) for z in grouper)
    desc = ('\n\nSome known OLD-STYLE British wedge designations:\n\n'
            'If you have one of those, then whenever the program asks you '
            'about the wedge designation,\nyou need to enter the NEW-STYLE '
            'number, for example "9 1/4 AK" will be "S5-9.25".\nIf the wedge '
            'is marked as "EU", then it was made for continental Europe,\n'
            'and based on old British pica value (.1667"). New-style wedge '
            'designations\nfor European wedges have the letter "E" '
            'appended to the set width - e.g. "S5-12E"\n')
    UI.display(desc)
    UI.display(old_wedges)
    UI.display('\n\nScroll your terminal up to see more.')


def display_wedge(wedge_name):
    """Display wedge's parameters"""
    try:
        wedge = Wedge(wedge_name)
        parameters = OrderedDict({False: 'Wedge {}'.format(wedge_name)})
        for key, value in wedge.parameters.items():
            parameters[key] = value
        UI.display_parameters(parameters)
    except ValueError:
        UI.display('Unknown wedge: {}'.format(wedge_name))


# Unit arrangement views


def list_unit_arrangements():
    """List all unit arrangements the program knows about."""
    number = 1
    # display UA list
    for ua_number in sorted(UNIT_ARRANGEMENTS, key=int):
        arr = UnitArrangement(ua_number)
        variant_names = ', '.join(var.name for var in arr.variants)
        UI.display('UA {} - {}'.format(arr.number, variant_names))
        with suppress(ValueError):
            number = int(ua_number)

    missing = [str(x) for x in range(1, number)
               if str(x) not in UNIT_ARRANGEMENTS]
    UI.display('The program currently knows {} unit arrangements.\nAt least '
               '{} seem to be missing - it is unclear if they ever existed.'
               '\n\nScroll your terminal up to see more.'
               .format(len(UNIT_ARRANGEMENTS), len(missing)))


def display_ua_variant(ua_variant):
    """Show an unit arrangement by char and by units"""
    def display_by_units():
        """display chars grouped by unit value"""
        UI.display('Ordered by unit value:')
        for unit_value, chars in sorted(ua_variant.by_units.items()):
            char_string = ', '.join(sorted(chars))
            UI.display('\t{}:\t{}'.format(unit_value or 'Absent', char_string))
        UI.display()

    def display_letters():
        """display unit values for all letters and ligatures"""
        # define templates for lower+uppercase, lowercase only, uppercase only
        template = '\t{:<4}: {:>3}, \t\t{:<4}: {:>3}'
        lc_tmpl = '\t{:<4}: {:>3}'
        uc_tmpl = '\t\t\t{:<4}: {:>3}'

        UI.display('Ordered by character:')
        for lowercase in [*ascii_lowercase, *d.LIGATURES]:
            uppercase = lowercase.upper()
            with suppress(ValueError, IndexError):
                # display both lower and upper
                lower_units = ua_variant[lowercase]
                upper_units = ua_variant[uppercase]
                UI.display(template.format(lowercase, lower_units or 'absent',
                                           uppercase, upper_units or 'absent'))
                continue
            with suppress(ValueError, IndexError):
                # display lowercase only
                lower_units = ua_variant[lowercase]
                UI.display(lc_tmpl.format(lowercase, lower_units or 'absent'))
                continue
            with suppress(ValueError, IndexError):
                # display uppercase only
                upper_units = ua_variant[uppercase]
                UI.display(uc_tmpl.format(uppercase, upper_units or 'absent'))
                continue
        UI.display()

    def display_numbers():
        """display 0...9 unit values"""
        grouped_by_units = OrderedDict()
        for units in range(7, 15):
            chars = []
            for number in digits:
                with suppress(ValueError):
                    if ua_variant[number] == units:
                        chars.append(number)
            if chars:
                grouped_by_units[units] = ', '.join(chars)
        if not grouped_by_units:
            return
        UI.display('Digits / numerals:')
        for units, characters in grouped_by_units.items():
            UI.display('{}: {} units'.format(characters, units))

    def display_others():
        """display other characters - not letters"""
        done = [*ascii_lowercase, *ascii_uppercase, *digits, *d.LIGATURES]
        other_chars = {u: [c for c in sorted(set(chars).difference(done))]
                       for u, chars in ua_variant.by_units.items()}
        others = [(', '.join(chars), units)
                  for units, chars in sorted(other_chars.items()) if chars]
        if others:
            chunks = ('{}: {} units'.format(c, u) if u
                      else '{}: absent'.format(c) for c, u in others)
            row = 'Other: {}'.format(', '.join(chunks))
            UI.display(row)

    if not ua_variant:
        UI.display('No unit arrangement to display...')
        return

    UI.display_header('{}'.format(ua_variant))
    display_by_units()
    display_letters()
    display_numbers()
    display_others()


def display_ua(ua_number):
    """Displays all variants in the unit arrangement."""
    arrangement = UnitArrangement(ua_number)
    for variant in arrangement.variants:
        display_ua_variant(variant)
        UI.display('\n')
    UI.display('Scroll your terminal up to see more.')


# Diecase views


def list_diecases(data):
    """Display all diecases in a dictionary, plus an empty new one"""
    UI.display('\nAvailable diecases:\n')
    UI.display_header('|{:<5}  {:<25} {:<12} {:<60}|'
                      .format('Index', 'Diecase ID', 'Wedge', 'Typeface(s)'),
                      trailing_newline=0)
    # show the rest of the table
    template = ('|{index:>5}  {d.diecase_id:<25} '
                '{d.wedge.name:<12} {d.description:<60}|')
    entries = (template.format(d=diecase, index=index)
               for index, diecase in data.items())
    UI.display(*entries, sep='\n')
    return data


def display_layout(diecase):
    """Display the diecase layout, unit values, styles."""
    def get_formatted_text(text, styles, length='^3'):
        """format a text with formatting string in styles definitions"""
        field = '{{:{}}}'.format(length) if length else '{}'
        collection = bm.Styles(styles)
        if text in d.SPACE_NAMES:
            # got a space = display symbol instead
            return field.format(d.SPACE_SYMBOLS.get(text[0]))
        elif collection.use_all:
            return field.format(text)
        else:
            ansi_codes = ';'.join(str(s.ansi) for s in collection if s.ansi)
            template = '\x1b[{codes}m{text}\x1b[0m'
            return template.format(codes=ansi_codes, text=field.format(text))

    def get_column_widths():
        """calculate column widths to adjust the content"""
        # 3 characters to start is reasonable enough
        columns = diecase.column_numbers
        column_widths = OrderedDict((name, 3) for name in columns)

        # get the maximum width of characters in every column
        # if it's larger than header field width, update the column width
        for column, initial_width in column_widths.items():
            widths_gen = (len(mat) for mat in diecase.select_column(column))
            column_widths[column] = max(initial_width, *widths_gen)
        return column_widths

    def build_row(row_number):
        """make a diecase layout table row - actually, 3 rows:
            empty row - separator; main row - characters,
            units row - unit values, if not matching the row units.
        """
        row = diecase.select_row(row_number)
        units = wedge.units[row_number]
        # initialize the row value dictionaries
        empty_row = dict(row='', units='')
        main_row = dict(row=row_number, units=units)
        units_row = dict(row='', units='')
        # fill all mat character fields
        for mat in row:
            column = mat.position.column
            empty_row[column] = ''
            # display unit width if it differs from row width
            m_units = '' if mat.units == units else mat.units
            units_row[column] = m_units
            # format character for display
            fmt = '^{}'.format(widths.get(column, 3))
            main_row[column] = get_formatted_text(mat.char, mat.styles, fmt)

        # format and concatenate two table rows
        empty_str = template.format(**empty_row)
        main_str = template.format(**main_row)
        units_str = template.format(**units_row)
        return '{}\n{}\n{}'.format(empty_str, main_str, units_str)

    def build_description():
        """diecase description: ID, typeface, wedge name and set width"""
        table_width = len(header)
        try:
            # diecase ID, typeface, wedge used
            row1_left = '{d.diecase_id} ({d.description})'.format(d=diecase)
            row1_right = 'wedge: {}'.format(wedge.name)
        except AttributeError:
            row1_left = 'Unknown diecase ID.'
            row1_right = 'Unknown wedge. Assuming S5.'
        row1_filled_width = len(row1_left) + len(row1_right) + 4
        row1_center = ' ' * (table_width - row1_filled_width)
        # available styles
        row2_left = ', '.join(get_formatted_text(s.name, s)
                              for s in diecase.styles)
        # warning: ANSI-formatting the strings affects their length!
        # calculate the correct length of formatted style names
        styles_length = len(', '.join(s.name for s in diecase.styles))
        # space symbols
        spaces = [(d.SPACE_NAMES.get(space), symbol)
                  for space, symbol in d.SPACE_SYMBOLS.items()]
        row2_right = ', '.join('{} = {}'.format(*item)
                               for item in sorted(spaces))
        # calculate the length of occupied space
        row2_filled_width = styles_length + len(row2_right) + 4
        row2_center = ' ' * (table_width - row2_filled_width)
        # table border template
        line = '═' * (table_width - 2)
        upper_border = '╔{}╗'.format(line)
        lower_border = '╠{}╣'.format(line)
        row_tmpl = '║ {}{}{} ║'
        return '\n'.join((upper_border,
                          row_tmpl.format(row1_left, row1_center, row1_right),
                          row_tmpl.format(row2_left, row2_center, row2_right),
                          lower_border))

    # what wedge does the diecase use?
    try:
        wedge = diecase.wedge
    except AttributeError:
        wedge = Wedge()
    # table row template
    widths = get_column_widths()
    fields = ' '.join(' {{{col}:^{width}}} '.format(col=col, width=width)
                      for col, width in widths.items())
    template = '║ {{row:>3}} │{fields}│ {{units:>5}} ║'.format(fields=fields)
    # header row template
    header = dict(units='Units', row='Row')
    header.update({col: col for col in widths})
    header = template.format(**header)
    # proper layout
    contents = (build_row(num) for num in diecase.row_numbers)
    # table description
    desc = build_description()
    # put the thing together
    table = (desc, header, '╟{}╢'.format('─' * (len(header) - 2)),
             *contents, '╚{}╝'.format('═' * (len(header) - 2)))
    # finally display it
    UI.display('\n'.join(table))
    UI.display('\n\nScroll your terminal up to see more.')


def display_diecase_data(diecase):
    """Show the diecase data, including typeface data"""
    typeface_data = diecase.typeface_data
    diecase_typefaces = [i for (i, _) in typeface_data.values()]
    for face in diecase_typefaces:
        calibration_line = '{:.4f}'.format(face.calibration_line)
        size = face.size
        if face and size:
            break
    else:
        face, size = 'unknown', 'unknown'

    header = 'Diecase {d.diecase_id} : {d.description}'.format(d=diecase)
    parameters = OrderedDict({False: header})
    parameters['Wedge'] = diecase.wedge
    parameters['Size'] = size
    parameters['Calibration line M'] = calibration_line

    UI.display_parameters(parameters)

    # display UAs typeface(s)
    arrangements = OrderedDict({False: 'Unit arrangements'})
    for style, (_, u_a) in typeface_data.items():
        arrangements[style.name] = '{ua.number} {ua.name}'.format(ua=u_a)
    UI.display_parameters(arrangements)


# Typeface views


def list_typefaces(*_):
    """Show all available typefaces"""
    template = '{series:>6}\t{name:<35}\t{script:<20} {styles:<25}\t{rel:<30}'
    typefaces = (Typeface(series) for series in sorted(TYPEFACES, key=int))
    fields = dict(series='Series', name='Name', script='Script',
                  styles='Styles in typeface', rel='Related typefaces')
    UI.display_header(template.format(**fields))
    # store the last number...
    number = 0
    # list of typefaces
    for typeface in typefaces:
        related = ', '.join('{}-> {} {}'
                            .format(style.name, face.series, face_style.name)
                            for style, (face, face_style)
                            in typeface.related.items())
        row = dict(series=typeface.series, name=typeface.name, rel=related,
                   script=typeface.script, styles=typeface.styles.names)
        UI.display(template.format(**row))
        with suppress(ValueError):
            number = int(typeface.series)

    # list missing typefaces
    missing = [str(x) for x in range(1, number) if str(x) not in TYPEFACES]
    UI.display('The program currently knows {} typefaces.\nAt least '
               '{} seem to be missing - it is unclear if they ever existed.'
               '\n\nScroll your terminal up to see more.'
               .format(len(TYPEFACES), len(missing)))


def display_typeface(typeface_id):
    """Get a typeface from the data and display its details:
    available styles, wedge set width, alignment M-line, classifiers
    """
    typeface = Typeface(typeface_id)
    related = ('{}-> {} {}'.format(style.name, face.series, face_style.name)
               for style, (face, face_style) in typeface.related.items())
    uas = ('{style.name}-> {ua.number} {ua.name}'.format(style=style, ua=ua)
           for style, ua in typeface.unit_arrangements.items())
    sizes = typeface.sizes

    # display general typeface data
    header = '{t.series} - {t.name}'.format(t=typeface)
    parameters = OrderedDict({False: header})
    parameters.update({'Classification': typeface.classification or 'unknown'})
    parameters.update({'Script': typeface.script or 'unknown'})
    parameters.update({'Related typefaces': ', '.join(related) or 'unknown'})
    parameters.update({'Unit arrangements': ', '.join(uas) or 'unknown'})
    UI.display_parameters(parameters)

    # display size-specific data if applicable
    if not sizes:
        return
    UI.display('\nTypeface size data:\n')
    template = '{size:<10} {sw:<10} {mline:<10} {uas}'
    UI.display(template.format(size='Size', sw='Set width', mline='M-line',
                               uas='Unit arrangement overrides'))
    for size, (set_width, m_line, uas) in sizes.items():
        size_arrangements = {bm.STYLES.get(st): UAVariant(ua_id, ua_v)
                             for st, (ua_id, ua_v) in uas.items()}
        arrs = ('{st.name}-> {ua.number} {ua.name}'.format(st=style, ua=ua)
                for style, ua in size_arrangements.items())
        UI.display(template.format(size=size, sw=set_width, mline=m_line,
                                   uas=', '.join(arrs)))
    # some explanation
    UI.display('\nWherever "@" appears in size, this means the matrices '
               'are cast on a different type body size\n'
               '(for example, 6D@7 means casting 6D mat on 7 point mould).\n\n'
               'M-line is a calibration line measured from the matrix '
               'upper edge.\n\n'
               'Some sizes (typically: large composition) can have '
               'different unit arrangements than general UAs\n'
               'for the typeface - they are listed as UA overrides.\n\n'
               'Scroll up your terminal to see more data.')



# -*- coding: utf-8 -*-
"""Caster object for either real or virtual Monotype composition caster"""
# Standard library imports
from collections import OrderedDict
from functools import wraps
from json.decoder import JSONDecodeError
import time

import requests
import librpi2caster
from librpi2caster import ON, OFF, CASTING, PUNCHING, HMN, KMN, UNITSHIFT

# Intra-package imports
from .rpi2caster import UI, Abort, option
from .parsing import parse_record


def handle_communication_error(routine):
    """If something goes wrong with a connection, ask what to do."""
    @wraps(routine)
    def wrapper(*args, **kwargs):
        """wrapper function"""
        try:
            return routine(*args, **kwargs)
        except librpi2caster.CommunicationError as error:
            message = ('Connection to the interface lost. Check your link.\n'
                       'Original error message:\n{}\n'
                       'Do you want to try again, or abort?')
            if UI.confirm(message.format(error), abort_answer=False):
                return routine(*args, **kwargs)
    return wrapper


class SimulationCaster:
    """Common methods for a caster class."""
    def __init__(self, url=None, operation_mode=None):
        self.url = url
        self.config, self.status = dict(), dict()
        self._setup()
        self.operation_mode = operation_mode

    def _setup(self):
        """Setup a simulation caster"""
        self.config = dict(supported_operation_modes=[CASTING, PUNCHING],
                           supported_row16_modes=[HMN, KMN, UNITSHIFT],
                           default_operation_mode=CASTING,
                           punching_on_time=0.2, punching_off_time=0.3,
                           sensor_timeout=5, pump_stop_timeout=20,
                           startup_timeout=20, name='Simulation interface')
        self.status = dict(water=OFF, air=OFF, motor=OFF, pump=OFF,
                           wedge_0005=15, wedge_0075=15, speed='0rpm',
                           working=OFF, testing_mode=OFF, signals=[],
                           current_operation_mode=CASTING,
                           current_row16_mode=OFF)

    def __str__(self):
        return self.config['name']

    def __call__(self, operation_mode=None, testing_mode=None):
        if operation_mode is not None:
            self.operation_mode = operation_mode
        if testing_mode is not None:
            self.testing_mode = testing_mode
        return self

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_):
        self.stop()

    def is_casting(self):
        """Determines if this caster is set to casting"""
        return self.operation_mode == CASTING

    @property
    def operation_mode(self):
        """Get the current operation mode:
            None = testing; casting; punching"""
        return self.status['current_operation_mode']

    @operation_mode.setter
    def operation_mode(self, mode):
        """Set the operation mode"""
        if self.status['working']:
            raise librpi2caster.InterfaceBusy
        elif not mode:
            default_operation_mode = self.config['default_operation_mode']
            self.status['current_operation_mode'] = default_operation_mode
        elif mode in self.supported_operation_modes:
            self.status['current_operation_mode'] = mode
        else:
            raise librpi2caster.UnsupportedMode

    @property
    def row16_mode(self):
        """Get the row 16 addressing mode:
            None = off, HMN, KMN, unit shift"""
        return self.status['current_row16_mode']

    @row16_mode.setter
    def row16_mode(self, mode):
        """Change the row 16 addressing mode:
            None (row 16 addressing is off = cast from row 15 instead)
            is available in all operation modes;

            HMN, KMN, unit shift are always available in the punching
            or testing modes, and conditionally available in the casting
            mode: they need to be supported in the interface configuration.
        """
        if self.status['working']:
            raise librpi2caster.InterfaceBusy
        elif not mode:
            self.status['current_row16_mode'] = OFF
        if mode not in [HMN, KMN, UNITSHIFT]:
            raise librpi2caster.UnsupportedRow16Mode
        elif self.is_casting() and mode not in self.supported_row16_modes:
            raise librpi2caster.UnsupportedRow16Mode
        else:
            self.status['current_row16_mode'] = mode

    @property
    def testing_mode(self):
        """Check whether the interface is in the testing mode"""
        return self.status['testing_mode']

    @testing_mode.setter
    def testing_mode(self, state):
        """Set the testing mode"""
        self.status['testing_mode'] = True if state else False

    @property
    def supported_operation_modes(self):
        """Get a list of supported operation modes"""
        return self.config['supported_operation_modes']

    @property
    def supported_row16_modes(self):
        """Get a list of supported row 16 addressing modes"""
        return self.config['supported_row16_modes']

    @property
    def parameters(self):
        """Gets a list of parameters"""
        parameters = OrderedDict({'': 'Caster parameters'})
        parameters.update(**self.config)
        return parameters

    def switch_operation_mode(self):
        """Switch between casting and punching"""
        new_mode = PUNCHING if self.is_casting() else CASTING
        try:
            self.operation_mode = new_mode
        except librpi2caster.UnsupportedMode:
            UI.pause('This caster does not support the {} mode'
                     .format(new_mode))
        except librpi2caster.InterfaceBusy:
            UI.pause('Cannot change the mode while the machine is working.')

    def choose_row16_mode(self, row16_needed=False):
        """Choose the addressing system for the diecase row 16."""
        if row16_needed and not self.row16_mode:
            # choose the row 16 addressing mode
            prompt = 'Choose the row 16 addressing mode'
            options = [option(key=name[0], value=name, text=name)
                       for name in self.supported_row16_modes]
            options.append(option(key='o', value=OFF,
                                  text='Off - cast from row 15 instead'))
            if len(options) > 1:
                self.row16_mode = UI.simple_menu(prompt, options,
                                                 default_key='o',
                                                 allow_abort=True)
            else:
                self.row16_mode = OFF

        elif self.row16_mode and not row16_needed:
            # turn off the unneeded row 16 adddressing attachment
            UI.display('The {} attachment is not needed - turn it off.'
                       .format(self.row16_mode))
            self.row16_mode = OFF

        else:
            if not self.row16_mode:
                UI.display('No row 16 addressing attachment is needed.')
            else:
                UI.display('Keep the {} attachment ON.'
                           .format(self.row16_mode))

    @property
    def signals(self):
        """Get the signals from the machine"""
        return self.status.get('signals', [])

    def start(self):
        """Simulates the machine start"""
        self.status.update(air=ON)
        if self.is_casting() and not self.testing_mode:
            self.status.update(water=ON, motor=ON)
        self.status.update(working=ON)

    def stop(self):
        """Simulates the machine stop"""
        if self.is_casting() and not self.testing_mode:
            self.status.update(pump=OFF, water=OFF, motor=OFF)
        self.status.update(air=OFF, working=OFF)
        self.testing_mode = False

    def send(self, signals, repeat=1, timeout=None, request_timeout=None):
        """Simulates sending the signals to the caster."""
        def update_wedges_and_pump():
            """Check the pump state and wedge positions;
            update the current status if neeeded."""
            def found(code):
                """check if code was found in a combination"""
                return set(code).issubset(signals)
            # check 0075 wedge position and determine the pump status:
            # find the earliest row number or default to 15
            if found(['0075']) or found('NK'):
                # 0075 always turns the pump on
                self.status['pump'] = ON
                for pos in range(1, 15):
                    if str(pos) in signals:
                        self.status['wedge_0075'] = pos
                        break
                else:
                    self.status['wedge_0075'] = 15

            elif found(['0005']) or found('NJ'):
                # 0005 without 0075 turns the pump off
                self.status['pump'] = OFF

            # check 0005 wedge position:
            # find the earliest row number or default to 15
            if found(['0005']) or found('NJ'):
                for pos in range(1, 15):
                    if str(pos) in signals:
                        self.status['wedge_0005'] = pos
                        break
                else:
                    self.status['wedge_0005'] = 15

        if not self.status['working']:
            raise librpi2caster.InterfaceNotStarted

        start_time = time.time()
        max_wait_time = max(timeout or 0, request_timeout or 0) or 10
        # convert signals based on modes
        parse = librpi2caster.parse_signals
        codes = parse(signals, self.operation_mode,
                      self.row16_mode, self.testing_mode)
        self.status['signals'] = codes
        try:
            for _ in range(repeat):
                update_wedges_and_pump()
                if self.testing_mode:
                    UI.display('Testing signals: {}'.format(''.join(codes)))
                elif self.is_casting():
                    UI.display('photocell ON')
                    time.sleep(0.5)
                    UI.display('photocell OFF')
                    time.sleep(0.5)
                else:
                    UI.display('punches going up')
                    time.sleep(self.config['punching_on_time'])
                    UI.display('punches going down')
                    time.sleep(self.config['punching_off_time'])
                if (time.time() - start_time) > max_wait_time:
                    raise librpi2caster.MachineStopped
        except (KeyboardInterrupt, EOFError):
            raise librpi2caster.MachineStopped
        return self.status

    def cast_one(self, record, timeout=None):
        """Casting sequence: sensor on - valves on - sensor off - valves off"""
        try:
            signals = record.signals
        except AttributeError:
            signals = record
        # calculate the timeout to prevent the request
        # from hanging indefinitely
        if timeout:
            request_timeout = 2 * timeout + 5
        else:
            request_timeout = 2 * self.config['sensor_timeout'] + 2
        return self.send(signals, timeout=timeout,
                         request_timeout=request_timeout)

    def punch_one(self, signals):
        """Punch a single signals combination"""
        off_time = self.config['punching_off_time']
        on_time = self.config['punching_on_time']
        timeout = on_time + off_time + 5
        return self.send(signals, request_timeout=timeout)

    def test_one(self, signals):
        """Test a single signals combination"""
        return self.send(signals, request_timeout=5)

    def cast(self, input_sequence, ask=True, repetitions=1):
        """Cast a series of multiple records.
        This is a simplified routine for internal use;
        normally use Casting.cast_ribbon which provides more info"""
        source = [parse_record(item) for item in input_sequence]
        # do we need to cast from row 16 and choose an attachment?
        row16_in_use = any(record.uses_row_16 for record in source)
        self.choose_row16_mode(row16_in_use)
        # use caster context to check machine rotation and ensure
        # that no valves stay open after we're done
        with self(operation_mode=CASTING):
            # cast as many as initially ordered and ask to continue
            while repetitions > 0:
                try:
                    for record in source:
                        UI.display('{r.signals:<20}{r.comment}'
                                   .format(r=record))
                        codes = self.cast_one(record).get('signals', [])
                        UI.display('casting: {}'.format(' '.join(codes)))
                    # repetition successful
                    repetitions -= 1
                except librpi2caster.MachineStopped:
                    # repetition failed
                    if UI.confirm('Machine stopped: continue casting?',
                                  abort_answer=False):
                        # restart the machine
                        self.start()
                if ask and not repetitions and UI.confirm('Repeat?'):
                    repetitions += 1

    def punch(self, signals_sequence):
        """Punching sequence: valves on - wait - valves off- wait"""
        # do we need to use any row16 addressing attachment?
        source = [parse_record(item) for item in signals_sequence]
        row16_in_use = any(record.uses_row_16 for record in source)
        self.choose_row16_mode(row16_in_use)
        # start punching
        with self(operation_mode=PUNCHING):
            for record in source:
                try:
                    UI.display('{r.signals:<20}{r.comment}'.format(r=record))
                    signals = self.punch_one(record.signals).get('signals', [])
                    UI.display('punching: {}'.format(' '.join(signals)))
                except librpi2caster.MachineStopped:
                    if UI.confirm('Machine stopped - continue?'):
                        signals = (self.punch_one(record.signals)
                                   .get('signals', []))
                        UI.display('punching: {}'.format(' '.join(signals)))
                    else:
                        break
            else:
                UI.display('All codes successfully punched.')
            UI.pause('Punching finished. Take the ribbon off the tower.')

    def test(self, signals_sequence, duration=None):
        """Testing: advance manually or automatically"""
        source = [parse_record(item) for item in signals_sequence]
        with self(testing_mode=ON):
            while True:
                try:
                    for record in source:
                        signals = (self.test_one(record.signals)
                                   .get('signals', []))
                        UI.display('testing: {}'.format(' '.join(signals)))
                        if duration:
                            time.sleep(duration)
                        else:
                            UI.pause('Next?', allow_abort=True)
                except librpi2caster.MachineStopped:
                    UI.display('Testing stopped.')
                UI.confirm('Repeat?', abort_answer=False)

    def diagnostics(self):
        """Settings and alignment menu for servicing the caster"""
        def test_front_pinblock(*_):
            """Sends signals 1...14, one by one"""
            info = 'Testing the front pinblock - signals 1...14.'
            UI.pause(info, allow_abort=True)
            self.test([str(n) for n in range(1, 15)])

        def test_rear_pinblock(*_):
            """Sends NI, NL, A...N"""
            info = 'This will test the rear pinblock - NI, NL, A...N. '
            UI.pause(info, allow_abort=True)
            self.test(['NI', 'NL', *'ABCDEFGHIJKLMN'])

        def test_all(*_):
            """Tests all valves and composition caster's inputs in original
            Monotype order:
            NMLKJIHGFSED 0075 CBA 123456789 10 11 12 13 14 0005.
            """
            info = ('This will test all the air lines in the same order '
                    'as the holes on the paper tower: \n{}\n'
                    'MAKE SURE THE PUMP IS DISENGAGED.')
            signals = [*'ONMLKJIHGFSED', '0075', *'CBA',
                       *(str(x) for x in range(1, 15)), '0005', '15']
            UI.pause(info.format(' '.join(signals)), allow_abort=True)
            self.test(signals)

        def test_justification(*_):
            """Tests the 0075-S-0005"""
            info = 'This will test the justification pinblock: 0075, S, 0005.'
            UI.pause(info, allow_abort=True)
            self.test(['0075', 'S', '0005'])

        def test_any_code(*_):
            """Tests a user-specified combination of signals"""
            with self(testing_mode=ON):
                while True:
                    UI.display('Enter the signals to send to the caster, '
                               'or leave empty to return to menu: ')
                    signals = UI.enter('Signals?', default=Abort)
                    output_signals = self.test_one(signals).get('signals', [])
                    UI.display('Sending {}'.format(' '.join(output_signals)))

        def blow_all(*_):
            """Blow all signals for a short time; add NI, NL also"""
            info = 'Blowing air through all air pins on both pinblocks...'
            UI.pause(info, allow_abort=True)
            queue = ['NI', 'NL', 'A1', 'B2', 'C3', 'D4', 'E5', 'F6', 'G7',
                     'H8', 'I9', 'J10', 'K11', 'L12', 'M13', 'N14',
                     '0075 S', '0005 O 15']
            self.test(queue, duration=0.3)

        def calibrate_wedges(*_):
            """Allows to calibrate the justification wedges so that when you're
            casting a 9-unit character with the S-needle at 0075:3 and 0005:8
            (neutral position), the    width is the same.

            It works like this:
            1. 0075 - turn the pump on,
            2. cast 7 spaces from the specified matrix (default: G5),
            3. put the line to the galley & set 0005 to 8, 0075 to 3, pump on,
            4. cast 7 spaces with the S-needle from the same matrix,
            5. put the line to the galley, then 0005 to turn the pump off.
            """
            UI.display('Transfer wedge calibration:\n\n'
                       'This function will cast two lines of 5 spaces: '
                       'first: G5, second: GS5 with wedges at 3/8. \n'
                       'Adjust the 52D space transfer wedge '
                       'until the lengths are the same.\n')
            UI.confirm('Proceed?', default=True, abort_answer=False)
            # prepare casting sequence
            record, justified_record = 'G7', 'GS7'
            pump_start, pump_stop = 'NKS 0075 3', 'NJS 0005 8'
            line_out = 'NKJS 0005 0075 8'
            # start - 7 x G5 - line out - start - 7 x GS5 - line out - stop
            sequence = [pump_start, *[record] * 7, line_out, pump_start,
                        *[justified_record] * 7, line_out, pump_stop]
            self.cast(sequence)

        def test_row_16(*_):
            """Tests the row 16 addressing attachment (HMN, KMN, unit-shift).
            Casts from all matrices in 16th row.
            """
            UI.display('This will test the 16th row addressing.\n'
                       'If your caster has HMN, KMN or unit-shift attachment, '
                       'turn it on.\n')
            # build casting queue
            pump_start, pump_stop = 'NKS 0075', 'NJS 0005'
            line_out = 'NKJS 0005 0075'
            row = ['{}16'.format(col)
                   for col in ('NI', 'NL', *'ABCDEFGHIJKLMNO')]
            # test with actual casting or not?
            if UI.confirm('Use the pump? Y = cast the row, N = test codes.'):
                sequence = [pump_start, *row, line_out, pump_stop]
                self.cast(sequence)
            else:
                self.choose_row16_mode(row16_needed=True)
                self.test(row)

        def options():
            """Generate the menu options"""
            ret = [option(key='a', value=test_all, seq=1,
                          text='Test outputs',
                          desc='Test all the air outputs N...O15, one by one'),
                   option(key='f', value=test_front_pinblock, seq=2,
                          text='Test the front pin block',
                          desc='Test the pins 1...14'),
                   option(key='r', value=test_rear_pinblock, seq=2,
                          text='Test the rear pin block',
                          desc='Test the pins NI, NL, A...N, one by one'),
                   option(key='b', value=blow_all, seq=2,
                          text='Blow all air pins',
                          desc='Blow air into every pin for a short time'),
                   option(key='j', value=test_justification, seq=2,
                          text='Test the justification block',
                          desc='Test the pins for 0075, S and 0005'),
                   option(key='c', value=test_any_code, seq=1,
                          text='Send specified signal combination',
                          desc='Send the specified signals to the machine'),
                   option(key='w', value=calibrate_wedges, seq=4,
                          cond=self.is_casting,
                          text='Calibrate the 52D wedge',
                          desc=('Calibrate the space transfer wedge '
                                'for correct width')),
                   option(key='l', value=test_row_16, seq=5,
                          cond=self.is_casting,
                          text='Test the extended 16x17 diecase system',
                          desc=('Cast type from row 16 '
                                'with HMN, KMN or unit-shift'))]
            return ret

        header = 'Diagnostics and machine calibration menu:'
        catch_exceptions = (Abort, KeyboardInterrupt, EOFError,
                            librpi2caster.MachineStopped)
        # Keep displaying the menu and go back here after any method ends
        UI.dynamic_menu(options=options, header=header, func_args=(self,),
                        catch_exceptions=catch_exceptions)


class MonotypeCaster(SimulationCaster):
    """Methods common for Caster classes, used for instantiating
    caster driver objects (whether real hardware or mockup for simulation)."""
    def _setup(self):
        """Initialize the caster"""
        if not self.url:
            message = 'Interface URL not specified!'
            raise librpi2caster.ConfigurationError(message)
        self.config = self._request(request_timeout=(1, 1)).get('settings')

    def _request(self, path='', request_timeout=None,
                 method=requests.get, **kwargs):
        """Encode data with JSON and send it to self.url in a request."""
        errors = {exc.code: exc
                  for exc in (librpi2caster.MachineStopped,
                              librpi2caster.UnsupportedMode,
                              librpi2caster.UnsupportedRow16Mode,
                              librpi2caster.InterfaceBusy,
                              librpi2caster.InterfaceNotStarted)}
        url = '{}/{}'.format(self.url, path).strip('/')
        data = kwargs or {}
        try:
            # get the info from the server
            response = method(url, json=data, timeout=request_timeout)
            # raise any HTTP errors ASAP
            response.raise_for_status()
            # we're sure we have a proper 200 OK status code
            reply = response.json()
            if not reply.get('success'):
                # shit happened in the driver
                error_code = reply.get('error_code', 666)
                exception = errors.get(error_code)
                if exception:
                    raise exception
            # return the content without the success value
            reply.pop('success')
            return reply
        except requests.exceptions.InvalidSchema:
            # wrong URL
            msg = 'The URL: {} must be a http://... or https://... address.'
            raise librpi2caster.ConfigurationError(msg.format(self.url))
        except requests.HTTPError as error:
            if response.status_code == 501:
                raise NotImplementedError('{}: feature not supported by server'
                                          .format(url))
            # 400, 404, 503 etc.
            raise librpi2caster.CommunicationError(str(error))
        except requests.Timeout:
            msg = 'Connection to {} timed out.'
            raise librpi2caster.CommunicationError(msg.format(self.url))
        except JSONDecodeError:
            msg = 'Connection to {} returned incorrect data (expected: JSON).'
            raise librpi2caster.CommunicationError(msg.format(self.url))
        except requests.ConnectionError:
            msg = 'Error connecting to {}.'
            raise librpi2caster.CommunicationError(msg.format(self.url))

    @property
    def operation_mode(self):
        """Get the current operation mode:
            None = testing; casting; punching"""
        return self.status['current_operation_mode']

    @operation_mode.setter
    @handle_communication_error
    def operation_mode(self, mode):
        """Set the operation mode"""
        self._request('operation_mode', method=requests.post, mode=mode)

    @property
    def row16_mode(self):
        """Get the row 16 addressing mode:
            None = off, HMN, KMN, unit shift"""
        return self.status['current_row16_mode']

    @row16_mode.setter
    @handle_communication_error
    def row16_mode(self, mode):
        """Change the row 16 addressing mode:
            None (row 16 addressing is off = cast from row 15 instead)
            is available in all operation modes;

            HMN, KMN, unit shift are always available in the punching
            or testing modes, and conditionally available in the casting
            mode: they need to be supported in the interface configuration.
        """
        self._request('row16_mode', method=requests.post, mode=mode)

    @property
    def testing_mode(self):
        """Get the current testing mode state"""
        return self.status['testing_mode']

    @testing_mode.setter
    @handle_communication_error
    def testing_mode(self, state):
        """Set the temporary testing mode on the interface"""
        value = True if state else False
        self._request(method=requests.post, testing_mode=value)

    @handle_communication_error
    def send(self, signals='', repeat=1, timeout=None, request_timeout=None):
        """Send signals to the interface. Wait for an OK response."""
        try:
            data = dict(signals=signals, repeat=repeat, timeout=timeout)
            return self._request('signals', method=requests.post,
                                 request_timeout=request_timeout, **data)
        except librpi2caster.InterfaceNotStarted:
            self.start()
            return self._request('signals', method=requests.post,
                                 request_timeout=request_timeout, **data)
        except (KeyboardInterrupt, EOFError, librpi2caster.CommunicationError):
            self.stop()
            raise librpi2caster.MachineStopped

    @handle_communication_error
    def start(self):
        """Machine startup sequence.
        In the casting mode:
            The interface will start the subsystems, if possible:
                compressed air supply, cooling water supply and motor.
            The operator has to turn the machine's shaft clutch on.
            The interface driver will detect rotation, and if the machine
            is stalling, a MachineStopped exception is raised.

        In other modes (testing, punching, manual punching):
            The interface will just turn on the compressed air supply.
        """
        if self.testing_mode:
            request_timeout = 5
        elif self.is_casting():
            info = ('Starting the composition caster...\n'
                    'Turn on the motor if necessary, and engage the clutch.\n'
                    'Casting will begin after detecting the machine rotation.'
                    '\n\nCommence casting? (N = abort)')
            if not UI.confirm(info):
                raise Abort
            request_timeout = 3 * self.config['startup_timeout'] + 2
        else:
            UI.pause('Waiting for you to start punching...')
            request_timeout = 5
        # send the request and handle any exceptions
        try:
            self._request('machine', method=requests.post,
                          request_timeout=request_timeout, machine=ON)
        except librpi2caster.InterfaceBusy:
            UI.pause('This interface is already working. Aborting...')
            raise Abort
        if self.is_casting():
            UI.display('OK, the machine is running...')

    @handle_communication_error
    def stop(self):
        """Machine stop sequence.
        The interface driver checks if the pump is active
        and turns it off if necessary.

        In the casting mode, the driver will turn off the motor
        and cut off the cooling water supply.
        Then, the air supply is cut off.
        """
        if self.testing_mode:
            request_timeout = 3
        elif self.is_casting():
            UI.display('Turning the machine off...')
            request_timeout = self.config['pump_stop_timeout'] * 2 + 2
        else:
            request_timeout = 3
        self._request('machine', method=requests.delete,
                      request_timeout=request_timeout)

    @property
    @handle_communication_error
    def status(self):
        """Get the interface status"""
        return self._request().get('status')

    @status.setter
    def status(self, _):
        """Prevents raising AttributeError when assigning to status"""
